#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module is to remove cosmic ray spaxels in the SEDM datacube.
-v20220705: change 'pd.append' > 'pd.concat'. And 'is'> '=='.
-v20201218-1221: add "show_cr_spaxels".
-v20201110.
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as mpl
from shapely.geometry import Point

from . import sedm
from . import io

def get_cr_spaxels_from_byecr(date, targetid):
    """
    Parameters
    ----------
    date: str
        YYYYMMDD
    targetid: str
        unique part of the filename, i.e. ZTF name or timestamp like 06_10_45

    Return
    ----------
    SEDC_BYECR class
    """

    return SEDM_BYECR.from_sedmid(date, targetid)

####################
#                  #
#   CLASS          #
#                  #
####################

class SEDM_BYECR():

    def __init__(self, date, cube):
        """
        set cube, hexagrid, and normalized cube.
        Then, derived_df also defines with spatial neighbors indexes.
        """

        self.cube = cube

        self.set_hexagrid(date)
        self.set_normalized_cube()

        self.derived_df = self.get_spatial_neighbors_indexes().reset_index()


    @classmethod
    def from_sedmid(cls, date, targetid):
        """
        """

        filename = io.get_night_files(date, "cube", targetid)[0]
        cube = sedm.get_sedmcube(filename)

        return cls(date, cube)


    #
    # SETTER
    #

    def set_hexagrid(self, date):
        """ NEED date_HexaGrid.pkl file. """
        self.hexagrid = io.load_nightly_hexagonalgrid(date)

    def set_normalized_cube(self):
        """
        Normalize every spaxel in a slice with same spaxels in -10 +10 slices.

        Returns
        -------
        normalized cube data and error (not variance).
        """

        #normalized_cube_data = np.empty( [self.cube.data.shape[0], self.cube.data.shape[1]])
        #normalized_cube_err = np.empty( [self.cube.data.shape[0], self.cube.data.shape[1]])

        cube_data_df = pd.DataFrame(self.cube.data, index=['lbda'+str(i) for i in range(0, self.cube.data.shape[0])]).T
        cube_var_df = pd.DataFrame(self.cube.variance, index=['lbda'+str(i) for i in range(0, self.cube.data.shape[0])]).T
        _lbda_num = len(self.cube.lbda)

        norm_cube_data_df = cube_data_df.copy()
        norm_cube_err_df = cube_var_df.copy()

        for _lbda_index in range(0, _lbda_num):

            if _lbda_index < 10:
                _median = np.median(cube_data_df.iloc[:, 0:_lbda_index+11], axis=1)
                norm_cube_data_df["lbda"+str(_lbda_index)] = cube_data_df.iloc[:, _lbda_index] / _median
                norm_cube_err_df["lbda"+str(_lbda_index)] = np.sqrt(cube_var_df.iloc[:, _lbda_index]) / _median
            elif 209 < _lbda_index:
                _median = np.median(cube_data_df.iloc[:, _lbda_index-10:_lbda_num], axis=1)
                norm_cube_data_df["lbda"+str(_lbda_index)] = cube_data_df.iloc[:, _lbda_index] / _median
                norm_cube_err_df["lbda"+str(_lbda_index)] = np.sqrt(cube_var_df.iloc[:, _lbda_index]) / _median
            else:
                _median = np.median(cube_data_df.iloc[:, _lbda_index-10:_lbda_index+11], axis=1)
                norm_cube_data_df["lbda"+str(_lbda_index)] = cube_data_df.iloc[:, _lbda_index] / _median
                norm_cube_err_df["lbda"+str(_lbda_index)] = np.sqrt(cube_var_df.iloc[:, _lbda_index]) / _median

        self.norm_cube_data = norm_cube_data_df.T.values
        self.norm_cube_err = norm_cube_err_df.T.values

        from pyifu import spectroscopy

        self.norm_cube = spectroscopy.get_cube(data=self.norm_cube_data, variance=self.norm_cube_err**2,
                                               header=self.cube.header, lbda=self.cube.lbda,
                                               spaxel_mapping=self.cube.spaxel_mapping,
                                               spaxel_vertices=self.cube.spaxel_vertices)


    #
    # GETTER
    #

    def get_spatial_neighbors_indexes(self):
        """ Only consider a spaxel witch has 6 neighbors to remove a spaxel at the edge of SEDM datacube."""

        _p_list = []

        for _test_index in range(0, self.cube.nspaxels):

            _neighbors_indexes = self.hexagrid.get_idx_neighbors(_test_index)

            if len(_neighbors_indexes) == 6:
                _neighbors_indexes_dict = {f"nei{i+1}": v for i, v in enumerate(_neighbors_indexes)}
                #_p_list.append(pd.DataFrame({"test_spx":_test_index, **_neighbors_indexes_dict}, index=[_test_index] ))
                _p_list.append(pd.DataFrame({**_neighbors_indexes_dict}, index=[_test_index] ))

        return pd.concat(_p_list, sort=True)

    def get_spatial_neighbors_info(self, lbda_index=0):
        """ """
        #print("Run get_spatial_neighbors_info")

        self.derived_df["flux1_norm"] = self.norm_cube_data[lbda_index][self.derived_df["nei1"]]
        self.derived_df["flux1_norm_err"] = self.norm_cube_err[lbda_index][self.derived_df["nei1"]]
        self.derived_df["flux2_norm"] = self.norm_cube_data[lbda_index][self.derived_df["nei2"]]
        self.derived_df["flux2_norm_err"] = self.norm_cube_err[lbda_index][self.derived_df["nei2"]]
        self.derived_df["flux3_norm"] = self.norm_cube_data[lbda_index][self.derived_df["nei3"]]
        self.derived_df["flux3_norm_err"] = self.norm_cube_err[lbda_index][self.derived_df["nei3"]]
        self.derived_df["flux4_norm"] = self.norm_cube_data[lbda_index][self.derived_df["nei4"]]
        self.derived_df["flux4_norm_err"] = self.norm_cube_err[lbda_index][self.derived_df["nei4"]]
        self.derived_df["flux5_norm"] = self.norm_cube_data[lbda_index][self.derived_df["nei5"]]
        self.derived_df["flux5_norm_err"] = self.norm_cube_err[lbda_index][self.derived_df["nei5"]]
        self.derived_df["flux6_norm"] = self.norm_cube_data[lbda_index][self.derived_df["nei6"]]
        self.derived_df["flux6_norm_err"] = self.norm_cube_err[lbda_index][self.derived_df["nei6"]]

        #self.derived_df["flux_norm_median"] = np.median(derived_df[["flux1_norm", "flux2_norm", "flux3_norm", "flux4_norm", "flux5_norm", "flux6_norm"]], axis=1)
        #self.derived_df["flux_norm_mad"] = stats.median_absolute_deviation(derived_df[["flux1_norm", "flux2_norm", "flux3_norm", "flux4_norm", "flux5_norm", "flux6_norm"]], axis=1)

        self.derived_df["nei_norm_mean"] = np.mean(self.derived_df[["flux1_norm", "flux2_norm", "flux3_norm", "flux4_norm", "flux5_norm", "flux6_norm"]], axis=1)
        self.derived_df["nei_norm_std"] = np.std(self.derived_df[["flux1_norm", "flux2_norm", "flux3_norm", "flux4_norm", "flux5_norm", "flux6_norm"]], axis=1)
        #self.derived_df["nei_norm_std_sqrtn"] = derived_df["nei_norm_std"] / np.sqrt(6) # number of neighbors = 6
        self.derived_df["nei_norm_mean_err"] = (1/6)*np.sqrt(self.derived_df["flux1_norm_err"]**2 +
                                                             self.derived_df["flux2_norm_err"]**2 +
                                                             self.derived_df["flux3_norm_err"]**2 +
                                                             self.derived_df["flux4_norm_err"]**2 +
                                                             self.derived_df["flux5_norm_err"]**2 +
                                                             self.derived_df["flux6_norm_err"]**2)


        return self.derived_df

    def get_spectral_neighbors_info(self, lbda_index=0):
        """ """

        #if lbda_index is 0:
        if lbda_index == 0:
            self.derived_df["spec_nei_flux1_norm"] = self.norm_cube_data[lbda_index+1][self.derived_df["index"]]
            self.derived_df["spec_nei_flux1_norm_err"] = self.norm_cube_err[lbda_index+1][self.derived_df["index"]]
            self.derived_df["spec_nei_flux2_norm"] = self.norm_cube_data[lbda_index+2][self.derived_df["index"]]
            self.derived_df["spec_nei_flux2_norm_err"] = self.norm_cube_err[lbda_index+2][self.derived_df["index"]]

        #elif lbda_index is (len(self.cube.lbda)-1):
        elif lbda_index == (len(self.cube.lbda)-1):
            self.derived_df["spec_nei_flux1_norm"] = self.norm_cube_data[lbda_index-1][self.derived_df["index"]]
            self.derived_df["spec_nei_flux1_norm_err"] = self.norm_cube_err[lbda_index-1][self.derived_df["index"]]
            self.derived_df["spec_nei_flux2_norm"] = self.norm_cube_data[lbda_index-2][self.derived_df["index"]]
            self.derived_df["spec_nei_flux2_norm_err"] = self.norm_cube_err[lbda_index-2][self.derived_df["index"]]

        else:
            self.derived_df["spec_nei_flux1_norm"] = self.norm_cube_data[lbda_index-1][self.derived_df["index"]]
            self.derived_df["spec_nei_flux1_norm_err"] = self.norm_cube_err[lbda_index-1][self.derived_df["index"]]
            self.derived_df["spec_nei_flux2_norm"] = self.norm_cube_data[lbda_index+1][self.derived_df["index"]]
            self.derived_df["spec_nei_flux2_norm_err"] = self.norm_cube_err[lbda_index+1][self.derived_df["index"]]

        self.derived_df["spec_nei_mean"] = np.mean(self.derived_df[["spec_nei_flux1_norm", "spec_nei_flux2_norm"]], axis=1)
        self.derived_df["spec_nei_std"] = np.std(self.derived_df[["spec_nei_flux1_norm", "spec_nei_flux2_norm"]], axis=1)
        self.derived_df["spec_nei_mean_err"] = (1/2)*np.sqrt(self.derived_df["spec_nei_flux1_norm_err"]**2 +
                                                             self.derived_df["spec_nei_flux2_norm_err"]**2)

        return self.derived_df

    def get_test_spaxel_info(self, lbda_index=0, wspectral=True):
        """ """

        self.derived_df["test_spaxel_flux_norm"] = self.norm_cube_data[lbda_index][self.derived_df["index"]]#.byteswap().newbyteorder()
        self.derived_df["test_spaxel_flux_norm_err"] = self.norm_cube_err[lbda_index][self.derived_df["index"]]#.byteswap().newbyteorder()

        self.derived_df["test_spaxel_norm_diff"] = self.derived_df["test_spaxel_flux_norm"] - self.derived_df["nei_norm_mean"]
        self.derived_df["test_spaxel_norm_diff_err"] = np.sqrt(self.derived_df["test_spaxel_flux_norm_err"]**2 + self.derived_df["nei_norm_mean_err"]**2)
        self.derived_df["test_spaxel_norm_diff_sigma"] = self.derived_df["test_spaxel_norm_diff"] / self.derived_df["test_spaxel_norm_diff_err"]

        if wspectral:
            self.derived_df["test_spaxel_spec_norm_diff"] = self.derived_df["test_spaxel_flux_norm"] - self.derived_df["spec_nei_mean"]
            self.derived_df["test_spaxel_spec_norm_diff_err"] = np.sqrt(self.derived_df["test_spaxel_flux_norm_err"]**2 + self.derived_df["spec_nei_mean_err"]**2)
            self.derived_df["test_spaxel_spec_norm_diff_sigma"] = self.derived_df["test_spaxel_spec_norm_diff"] / self.derived_df["test_spaxel_spec_norm_diff_err"]

        return self.derived_df

    #def get_cr_spaxel_info(self, lbda_index=0, wspectral=False, cut_criteria=5):
    def get_cr_spaxel_info(self, lbda_index=None, wspectral=False, cut_criteria=5):
        """
        Return cr spaxel information, e.g., wavelength, normalized flux, neighbors mean etc.

        Parameters
        ----------
        lbda_index: float.
            specify lbda_index you want to investigate.
            'None' means run through whole lbda_index. Default.
        wspectral: bool.
            if you want to use a spectral filtering together with spatial one, put 'wspectral=True'.
            But it is not yet validated.
        cut_criteria: float.
            cut criteria values we want to use.
            Default is '5'.

        Return
        ----------
        cosmic ray information DataFrame.
        """

        self.cut_criteria = cut_criteria

        cr_info_df = pd.DataFrame(columns=["cr_spaxel_index", "cr_spaxel_id", "cr_lbda", "cr_lbda_index", "cr_diff_norm_sigma",
                                           "test_spaxel_flux_norm", "test_spaxel_flux_norm_err",
                                           "nei_norm_mean", "nei_norm_mean_err"])

        if lbda_index is not None: #for specific lbda_index to test

            _lbda_index = lbda_index

            self.get_spatial_neighbors_info(lbda_index=_lbda_index)
            self.get_test_spaxel_info(lbda_index=_lbda_index, wspectral=False)

            _cr_info_df = pd.DataFrame(columns=["cr_spaxel_index", "cr_lbda", "cr_lbda_index", "cr_diff_norm_sigma",
                                                "test_spaxel_flux_norm", "test_spaxel_flux_norm_err",
                                                "nei_norm_mean", "nei_norm_mean_err"])

            _cr_temp_df = self.derived_df[ abs(self.derived_df["test_spaxel_norm_diff_sigma"]) > self.cut_criteria ]

            if len(_cr_temp_df) > 0:

                _cr_info_df["cr_spaxel_index"] = _cr_temp_df["index"]
                _cr_info_df["cr_spaxel_id"] = self.cube.indexes[_cr_info_df["cr_spaxel_index"]]
                _cr_info_df["cr_lbda"] = self.cube.lbda[_lbda_index]
                _cr_info_df["cr_lbda_index"] = _lbda_index
                _cr_info_df["cr_diff_norm_sigma"] = _cr_temp_df["test_spaxel_norm_diff_sigma"]
                _cr_info_df["test_spaxel_flux_norm"] = _cr_temp_df["test_spaxel_flux_norm"]
                _cr_info_df["test_spaxel_flux_norm_err"] = _cr_temp_df["test_spaxel_flux_norm_err"]
                _cr_info_df["nei_norm_mean"] = _cr_temp_df["nei_norm_mean"]
                _cr_info_df["nei_norm_mean_err"] = _cr_temp_df["nei_norm_mean_err"]

                #cr_info_df = cr_info_df.append( _cr_info_df, ignore_index=True)
                cr_info_df = pd.concat([cr_info_df, _cr_info_df], ignore_index=True)

        else: #run whole lbda_index
            for _lbda_index in range(0, len(self.cube.lbda)):

                if wspectral:
                    self.get_spatial_neighbors_info(lbda_index=_lbda_index)
                    self.get_spectral_neighbors_info(lbda_index=_lbda_index)
                    self.get_test_spaxel_info(lbda_index=_lbda_index, wspectral=True)

                    _cr_info_df = pd.DataFrame(columns=["cr_spaxel_index", "cr_spaxel_id", "cr_lbda", "cr_lbda_index", "cr_diff_norm_sigma",
                                                        "test_spaxel_flux_norm", "test_spaxel_flux_norm_err",
                                                        "nei_norm_mean", "nei_norm_mean_err"])

                    _cr_temp_df = self.derived_df[ abs(self.derived_df["test_spaxel_norm_diff_sigma"]) > self.cut_criteria ]

                    if len(_cr_temp_df) > 0:

                        _cr_info_df["cr_spaxel_index"] = _cr_temp_df["index"]
                        _cr_info_df["cr_spaxel_id"] = self.cube.indexes[_cr_info_df["cr_spaxel_index"]]
                        _cr_info_df["cr_lbda"] = self.cube.lbda[_lbda_index]
                        _cr_info_df["cr_lbda_index"] = _lbda_index
                        _cr_info_df["cr_diff_norm_sigma"] = _cr_temp_df["test_spaxel_norm_diff_sigma"]
                        _cr_info_df["test_spaxel_flux_norm"] = _cr_temp_df["test_spaxel_flux_norm"]
                        _cr_info_df["test_spaxel_flux_norm_err"] = _cr_temp_df["test_spaxel_flux_norm_err"]
                        _cr_info_df["nei_norm_mean"] = _cr_temp_df["nei_norm_mean"]
                        _cr_info_df["nei_norm_mean_err"] = _cr_temp_df["nei_norm_mean_err"]

                        #cr_info_df = cr_info_df.append( _cr_info_df, ignore_index=True)
                        cr_info_df = pd.concat([cr_info_df, _cr_info_df], ignore_index=True)

                else: # wspectral=False, default
                    self.get_spatial_neighbors_info(lbda_index=_lbda_index)
                    self.get_test_spaxel_info(lbda_index=_lbda_index, wspectral=False)

                    _cr_info_df = pd.DataFrame(columns=["cr_spaxel_index", "cr_spaxel_id", "cr_lbda", "cr_lbda_index", "cr_diff_norm_sigma",
                                                        "test_spaxel_flux_norm", "test_spaxel_flux_norm_err",
                                                        "nei_norm_mean", "nei_norm_mean_err"])

                    _cr_temp_df = self.derived_df[ abs(self.derived_df["test_spaxel_norm_diff_sigma"]) > self.cut_criteria ]

                    if len(_cr_temp_df) > 0:

                        _cr_info_df["cr_spaxel_index"] = _cr_temp_df["index"]
                        _cr_info_df["cr_spaxel_id"] = self.cube.indexes[_cr_info_df["cr_spaxel_index"]]
                        _cr_info_df["cr_lbda"] = self.cube.lbda[_lbda_index]
                        _cr_info_df["cr_lbda_index"] = _lbda_index
                        _cr_info_df["cr_diff_norm_sigma"] = _cr_temp_df["test_spaxel_norm_diff_sigma"]
                        _cr_info_df["test_spaxel_flux_norm"] = _cr_temp_df["test_spaxel_flux_norm"]
                        _cr_info_df["test_spaxel_flux_norm_err"] = _cr_temp_df["test_spaxel_flux_norm_err"]
                        _cr_info_df["nei_norm_mean"] = _cr_temp_df["nei_norm_mean"]
                        _cr_info_df["nei_norm_mean_err"] = _cr_temp_df["nei_norm_mean_err"]

                        #cr_info_df = cr_info_df.append( _cr_info_df, ignore_index=True)
                        cr_info_df = pd.concat([cr_info_df, _cr_info_df], ignore_index=True)



        dtype_dict = {"cr_spaxel_index": int, "cr_lbda": float, "cr_lbda_index": int, "cr_diff_norm_sigma": float,
                      "test_spaxel_flux_norm": float, "test_spaxel_flux_norm_err": float,
                      "nei_norm_mean": float, "nei_norm_mean_err": float}
        cr_info_df = cr_info_df.astype(dtype_dict)

        return cr_info_df.sort_values(["cr_spaxel_index", "cr_lbda"])

    #
    # SHOW
    #

    def show_cr_spaxels(self, wcube=False, lbda_index=None, wspectral=False, cut_criteria=5, savefile=None):
        """ """

        cr_df = self.get_cr_spaxel_info(lbda_index=lbda_index, wspectral=wspectral, cut_criteria=cut_criteria)

        fig = mpl.figure(figsize=[5,5])

        ax = fig.add_subplot(111)

        _text = "Number of detected cosmic rays = %i\nNumber of affected spaxels: %i" %( len(cr_df), len(np.unique(cr_df["cr_spaxel_index"])) )

        if wcube:
            spaxel_patches = self.cube._display_im_(ax, vmin="5", vmax="99")

            for i_cr in np.unique(cr_df["cr_spaxel_index"]):
                spaxel_patches[i_cr].set_edgecolor("blue")
                spaxel_patches[i_cr].set_linewidth(2)
                spaxel_patches[i_cr].set_zorder(13)

            ax.text(-20, 23.0, _text, fontsize=10)
        else: #only text
            ax.set_xlim(0, 10)
            ax.set_ylim(0, 10)
            ax.xaxis.set_visible(False)
            ax.yaxis.set_visible(False)

            ax.text(0.2, 5, _text, fontsize=13)


        if savefile is not None:
            fig.savefig( savefile )



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import os
import warnings
#import matplotlib
#matplotlib.use('Agg')
import matplotlib.pyplot as mpl
from glob import glob

from astrobject.utils.tools import dump_pkl
from astropy.io import fits

from .. import io

from ..ccd import get_ccd
from ..spectralmatching import get_tracematcher, illustrate_traces, load_trace_masks
from ..wavesolution import get_wavesolution, Flexure

from ..sedm import INDEX_CCD_CONTOURS, TRACE_DISPERSION, build_sedmcube, build_calibrated_sedmcube, SEDM_LBDA

############################
#                          #
#  Spectral Matcher        #
#                          #
############################
def build_tracematcher(date, verbose=True, width=None,
                           save_masks=False,
                           rebuild=False,
                           ncore=None):
    
    """ Create Spaxel trace Solution 
    This enable to know which pixel belong to which spaxel

    Parameters
    ----------
    date: [string]
        YYYMMDD 

    width: [None/float] -optional-
        What should be the trace width (in pixels)
        If None, this will use 2 times whatever is defined as "TRACE_DISPERSION" in sedm.py
        
    save_masks: [bool] -optional-
        Shall this measures all the individual masks and save them?
        This mask building is the most cpu expensive part of the pipeline, 
        it takes about 30min/number_of_core. If this is saved (~150Mo) 
        wavelength-solution and cube-building will be faster.
        
    Returns
    -------
    Void.  (Creates the file TraceMatch.pkl and TraceMatch_WithMasks.pkl if save_masks)
    """
    
    
    timedir = io.get_datapath(date)
        
    if verbose:
        print("Directory affected by Spectral Matcher : %s"%timedir)
        
    if width is None:
        width = TRACE_DISPERSION

    # - Load the Spectral Matcher
    if not rebuild:
        try:
            smap = io.load_nightly_tracematch(date)
        except:
            rebuild = True
        
    if rebuild:
        print("Building Nightly Solution")
        smap = get_tracematcher(glob(timedir+"dome.fits*")[0], width=width)
        smap.writeto(timedir+"%s_TraceMatch.pkl"%date)
        print("Nightly Solution Saved")
        
    if save_masks:
        if not rebuild and len(glob(timedir+"%s_TraceMatch_WithMasks.pkl"%date))>0:
            warnings.warn("TraceMatch_WithMasks already exists for %s. rebuild is False, so nothing is happening"%date)
            return
        load_trace_masks(smap, smap.get_traces_within_polygon(INDEX_CCD_CONTOURS), ncore=ncore)
        smap.writeto(timedir+"%s_TraceMatch_WithMasks.pkl"%date)
    
############################
#                          #
# Spaxel Spacial Position  #
#                          #
############################
def build_hexagonalgrid(date, xybounds=None, theta=None):
    """ """
    smap  = io.load_nightly_tracematch(date)
    # ----------------
    # - Spaxel Selection
    if xybounds is None:
        xybounds=INDEX_CCD_CONTOURS
    idxall = smap.get_traces_within_polygon(xybounds)

    hgrid = smap.extract_hexgrid(idxall, theta=theta)

    timedir = io.get_datapath(date)
    hgrid.writeto(timedir+"%s_HexaGrid.pkl"%date)

############################
#                          #
# Spaxel Spacial Position  #
#                          #
############################
def build_flatfield(date, lbda_min=7000, lbda_max=9000,
                    ref="dome", build_ref=True,
                    kind="median", savefig=True, ncore=None):
    """ """
    from ..sedm import get_sedmcube
    from pyifu.spectroscopy  import get_slice

    reffile  = io.get_night_files(date, kind="cube.basic", target=ref)

    # - If the reference if not there yet.
    if len(reffile)==0:
        warnings.warn("The reference cube %s does not exist "%ref)
        if build_ref:
            warnings.warn("build_flatfield is building it!")
        else:
            raise IOError("No reference cube to build the flatfield (build_ref was set to False)")
        # --------------------- #
        # Build the reference   #
        # --------------------- #
        tmatch   = io.load_nightly_tracematch(date, withmask=True) 
        # - The CCD
        ccdreffile = io.get_night_files(date, kind="ccd.lamp", target=ref)[0]
        ccdref     = get_ccd(ccdreffile, tracematch = tmatch, background = 0)
        ccdref.fetch_background(set_it=True, build_if_needed=True, ncore=ncore)
        if not ccdref.has_var():
            ccdref.set_default_variance()
        # - HexaGrid
        hgrid    = io.load_nightly_hexagonalgrid(date)
        wcol     = io.load_nightly_wavesolution(date)
        wcol._load_full_solutions_()
        # - Build a cube
        build_sedmcube(ccdref, date, lbda=None, wavesolution=wcol, hexagrid=hgrid,
                        flexure_corrected=False,
                        flatfielded=False, build_calibrated_cube=False,atmcorrected=False)
        
    # ---------------------- #
    #  Actual FlatFielding   #
    # ---------------------- #
    reffile  = reffile  = io.get_night_files(date, kind="cube.basic", target=ref)[0]
    refcube  = get_sedmcube(reffile)
    sliceref = refcube.get_slice(lbda_min=lbda_min, lbda_max=lbda_max, usemean=True)
    # - How to normalize the Flat
    if kind in ["med", "median"]:
        norm = np.nanmedian(sliceref)
    elif kind in ["mean"]:
        norm = np.nanmean(sliceref)
    elif kind in refcube.indexes:
        norm = sliceref[np.argwhere(refcube.indexes==kind)]
    else:
        raise ValueError("Unable to parse the given kind: %s"%kind)
    
    # - The Flat
    flat     = sliceref / norm
    slice_ = get_slice(flat, np.asarray(refcube.index_to_xy(refcube.indexes)),
                        refcube.spaxel_vertices,
                        indexes=refcube.indexes, variance=None, lbda=None)
    # - Figure
    timedir  = io.get_datapath(date)
    # - Saving
    slice_.header["CALTYPE"] = "FlatField"
    slice_.header["FLATSRC"]  = ref
    slice_.header["FLATREF"]  = kind
    slice_.writeto(timedir+'%s_Flat.fits'%date)
    
    if savefig:
        print("Saving flat3d Figure")
        slice_.show(savefile=timedir+"%s_flat3d.pdf"%date)
        slice_.show(savefile=timedir+"%s_flat3d.png"%date)

    
############################
#                          #
#   BackGround             #
#                          #
############################
def build_backgrounds(date, smoothing=[0,5], start=2, jump=10, 
                        target=None, lamps=True, only_lamps=False, skip_calib=True,
                        multiprocess=True, show_progress=False,
                        savefig=True, ncore=None):
    """ """
    from ..background import build_background
    timedir  = io.get_datapath(date)
    # - The files
    fileccds = []
    if not only_lamps:
        crrfiles  = io.get_night_files(date, "ccd.crr", target=target)
        if skip_calib: fileccds = [f for f in crrfiles if "Calib" not in fits.getval(f,"Name")]            
        fileccds += crrfiles

    # - Building the background
    tmap = io.load_nightly_tracematch(date)
    nfiles = len(fileccds)
    print("%d files to go..."%nfiles)
    for i,file_ in enumerate(fileccds):
        build_background(get_ccd(file_, tracematch=tmap, background=0),
                        start=start, jump=jump, multiprocess=multiprocess, 
                        smoothing=smoothing,show_progress=show_progress,
                        savefile = None if not savefig else timedir+"bkgd_%s.pdf"%(file_.split('/')[-1].replace(".fits","")),
                        ncore=ncore)
        
    
############################
#                          #
#  Wavelength Solution     #
#                          #
############################
def build_wavesolution(date, verbose=False, ntest=None, idxrange=None,
                       use_fine_tuned_traces=False,
                       wavedegree=5, contdegree=3, show_progress=False,
                       lamps=["Hg","Cd","Xe"], savefig=True, saveindividuals=False,
                       xybounds=None, rebuild=True):
    """ Create the wavelength solution for the given night.
    The core of the solution fitting is made in pysedm.wavesolution.

    Parameters
    ----------
    
    Returns
    -------

    """
    timedir = io.get_datapath(date)
        
    if verbose:
        print("Directory affected by Wavelength Calibration: %s"%timedir)


    if not rebuild and len(glob(timedir+"%s_WaveSolution.pkl"%(date)))>0:
        warnings.warn("WaveSolution already exists for %s. rebuild is False, so nothing is happening"%date)
        return
    
    # ----------------
    # - Load the Data
    # - SpectralMatch using domes
    #   Built by build_spectmatcher
    smap = io.load_nightly_tracematch(date, withmask=True)
        
    if use_fine_tuned_traces:
        raise ValueError("use_fine_tuned_traces is not supported anymore")
    
    fileccd_lamps = io.get_night_files(date, "ccd.lamp", target="|".join(lamps))
    lamps = [get_ccd(f_, tracematch=smap, correct_traceflexure=True,
                    savefile_traceflexure=None) # Don't save the figure here (called a lot when using derive_wavesolution)
                 for f_ in fileccd_lamps]
    
    # - The CubeSolution
    csolution = get_wavesolution(*lamps)

    # ----------------
    # - Spaxel Selection
    if xybounds is None:
        xybounds = INDEX_CCD_CONTOURS
        
    idxall = smap.get_traces_within_polygon(xybounds)
    if idxrange is not None:
        idxall = [l for l in idxall if l>=idxrange[0] and l<idxrange[1]]
    idx = idxall if ntest is None else np.random.choice(idxall,ntest, replace=False) 

    # - Do The loop and map it thanks to astropy
    def fitsolution(idx_):
        if saveindividuals:
            saveplot = timedir+"%s_wavesolution_trace%d.pdf"%(date,idx_)
        else:
            saveplot = None
        csolution.fit_wavelesolution(traceindex=idx_, saveplot=None,
                    contdegree=contdegree, wavedegree=wavedegree, plotprop={"show_guesses":True})
        if saveplot is not None:
            csolution._wsol.show(show_guesses=True, savefile=saveplot)
            mpl.close("all")

    if show_progress:
        from astropy.utils.console import ProgressBar
        from ..utils import tools
        notebook = tools.is_running_from_notebook()
        bar = ProgressBar( len(index_columnidx), ipython_widget=notebook)
    else:
        bar = None

    for j,i_ in enumerate(idx):
        fitsolution(i_)
        if bar is not None:
            bar.update(j)
    if bar is not None:
        bar.update(len(idx))
        
    # - output - #
    outfile = "%s_WaveSolution"%date
    if idxrange is not None:
        outfile += "_range_%d_%d"%(idxrange[0],idxrange[1])
    if ntest is not None:
        outfile += "_ntest%d"%(ntest)
        
    dump_pkl(csolution.wavesolutions, timedir+"%s.pkl"%outfile)
    
    if savefig:
        if ntest is not None or idxrange is not None:
            print("No WaveSolution saving plot when ntest or idxrange are set. ")
        else:
            print("Saving Wavesolution plot")
            wsol = io.load_nightly_wavesolution(date)
            hexagrid = io.load_nightly_hexagonalgrid(date)
            pl = wsol.show_dispersion_map(hexagrid,vmin="0.5",vmax="99.5",
                                              outlier_highlight=5, show=False)
            pl['fig'].savefig(timedir+"%s_wavesolution_dispersionmap.pdf"%date)
            pl['fig'].savefig(timedir+"%s_wavesolution_dispersionmap.png"%date)

        
        
    
############################
#                          #
#  Build Cubes             #
#                          #
############################
def build_night_cubes(date, target=None, lamps=True, only_lamps=False,
                          skip_calib=True, ncore=None, **kwargs):
    """ 
    """
    fileccds = []
    if lamps:
        fileccds += io.get_night_files(date, "ccd.lamp", target=target)
    if not only_lamps:
        crrfiles  = io.get_night_files(date, "ccd.crr", target=target)
        if skip_calib: crrfiles = [f for f in crrfiles if "Calib" not in fits.getval(f,"Name")]            
        fileccds += crrfiles

    print("fileccds:", fileccds)
    build_cubes(fileccds, date, ncore=ncore, **kwargs)


# ----------------- #
#  Build Cubes      #
# ----------------- #
def build_cubes(ccdfiles,  date, lbda=None,
                tracematch=None, wavesolution=None, hexagrid=None,
                # Background:
                nobackground = False,
                # Flat Field
                flatfielded=True, flatfield=None,
                # Flexure
                traceflexure_corrected=True, flexure_corrected=True,
                # Calibration
                atmcorrected=True, 
                build_calibrated_cube=False, calibration_ref=None,
                # Out
                build_guider=True, solve_wcs=False,
                fileindex=None, show_progress=False,
                savefig=True, verbose=True, 
                ncore=None):
    """ Build a cube from the an IFU ccd image. This image 
    should be bias corrected and cosmic ray corrected.

    The standard created cube will be 3Dflat field correct 
    (see flatfielded) and corrected for atmosphere extinction 
    (see atmcorrected). A second cube will be flux calibrated 
    if possible (see build_calibrated_cube).

    Parameters
    ----------
    ccd: [CCD]
        A ccd object from which the cube will be extracted.
        
    date: [string]
        date in usual YYYYMMDD format
    
    lbda: [None/array] -optional-
        wavelength array used to build the cube. 
        If not given the default sedm wavelength range will be used. 
        See pysedm.sedm.SEDM_LBDA.

    // Cube Calibrator //
    
    wavesolution: [WaveSolution] -optional-
        The wavelength solution containing the pixel<->wavelength conversion.
        If None, this will be loaded using `date`.

    hexagrid: [HexaGrid] -optional-
        The Hexagonal Grid tools containing the index<->qr<->xy conversions.
        If None, this will be loaded using `date`.

    flatfield: [Slice] -optional-
        Object containing the relative transmission of the IFU spaxels.
        If None, this will be loaded using `date`.

    // Action Selection //

    nobackground: [bool] -option-
        Shall the ccd background be 0 instead of the usual pipeline background?

    flexure_corrected: [bool] -optional-
        Shall the cube be flexure corrected ?
        - Remark, this means the cube will be built twice: 
            - 1 time to estimated the flexure
            - 1 time with flexure correction applied.

    flatfielded: [bool] -optional-
        Shall the cube be flatfielded?
        - This information will be saved in the header-
    
    atmcorrected: [bool] -optional-
        Shall the cube the corrected for atmosphere extinction?
        - This information will be saved in the header-

    // Additional outcome: Flux calibrated cube //

    build_calibrated_cube: [bool] -optional-
        Shall this method build an additionnal flux calibrated cube?
        
    calibration_ref: [None/string] -optional-
        If you want to build a calibrated cube, you can provide the filename
        of the spectrum containing the inverse-sensitivity (fluxcal*)
        If None, this will load the latest fluxcal object of the night.
        If Nothing found, no flux calibrated cube will be created. 

    Returns
    -------
    Void
    """
    if traceflexure_corrected:
        from ..flexure import get_ccd_jflexure
    # ------------------ #
    # Loading the Inputs #
    # ------------------ #
    if tracematch is None:
        tracematch   = io.load_nightly_tracematch(date, withmask=False if traceflexure_corrected else True) 
        
    if hexagrid is None:
        hexagrid     = io.load_nightly_hexagonalgrid(date)
    
    if wavesolution is None:
        wavesolution = io.load_nightly_wavesolution(date)
        wavesolution._load_full_solutions_()
    
    if lbda is None:
        lbda = SEDM_LBDA

    if flatfielded and flatfield is None:
        flatfield = io.load_nightly_flat(date)
        
    # ---------------- #
    # Loading the CCDS #
    # ---------------- #
    ccds = []
    for ccdfile in ccdfiles:
        flexuresavefile = None if not savefig else [ccdfile.replace("crr","flexuretrace_crr").replace(".fits",".pdf"),
                                                    ccdfile.replace("crr","flexuretrace_crr").replace(".fits",".png")]
        ccd_    = get_ccd(ccdfile, tracematch = tracematch, background = 0,
                              correct_traceflexure = traceflexure_corrected,
                              savefile_traceflexure=flexuresavefile)
        if traceflexure_corrected:
            load_trace_masks(ccd_.tracematch,
                                ccd_.tracematch.get_traces_within_polygon(INDEX_CCD_CONTOURS),
                                ncore=ncore)
            
        if not nobackground:
            ccd_.fetch_background(set_it=True, build_if_needed=True, ncore=ncore)
            ccd_.header["CCDBKGD"] = (True, "is the ccd been background subtracted?")
        else:
            ccd_.header["CCDBKGD"] = (False, "is the ccd been background subtracted?")
            
        # - Variance
        if not ccd_.has_var():
            ccd_.set_default_variance()
        ccds.append(ccd_)
            
    # ---------------- #
    # Build the Cubes  #
    # ---------------- #
    # internal routine 
    def _build_cubes_(ccdin):
        print(ccdin.filename)
        prop = dict(lbda=lbda, wavesolution=wavesolution, hexagrid=hexagrid,
                    flexure_corrected=flexure_corrected,
                    flatfielded=flatfielded, flatfield=flatfield,
                    atmcorrected=atmcorrected, 
                    build_calibrated_cube=build_calibrated_cube,
                    calibration_ref=calibration_ref,
                    fileindex=fileindex,
                    savefig=savefig)
        if build_guider:
            from pysedm import rainbowcam
            try:
                print("INFO: building the guider image")
                rainbowcam.build_meta_ifu_guider(ccdin.filename, solve_wcs=solve_wcs)
            except:
                print("WARNING: rainbowcam cannot build the guider image")

        build_sedmcube(ccdin, date,  **prop)
            
            
    # The actual build
    return [_build_cubes_(ccd_) for ccd_ in ccds]
        

# ---------------- #
# Flux Calibration #
# ---------------- #
def calibrate_night_cubes(date, target=None, calibrated_reference=None):
    """ """
    files = io.get_night_files(date, "cube.basic", target=target)
    flux_calibrate_cubes(files, date=date, calibrated_reference=calibrated_reference)
    

    
def calibrate_cubes(cubefiles, date=None, calibrated_reference=None,
                             multiprocess=True):
    """ """
    # internal routine 
    def _build_cal_cubes_(cubefile):
        build_calibrated_sedmcube(cubefile, date=date,
                                calibration_ref=calibrated_reference)
    # - The build
    if len(cubefiles)==1:
        return _build_cal_cubes_(cubefiles[0])
    if multiprocess:
        print("multiprocess removed for calibrate_cubes()")
        
    return [_build_cal_cubes_(cubefiles_) for cubefiles_ in cubefiles]
        
    
def save_cubeplot(date, kind="cube.basic"):
    """ """
    from ..sedm import get_sedmcube
    timedir  = io.get_datapath(date)
    
    for cubefile in io.get_night_files(date, kind):
        cube = get_sedmcube(cubefile)
        cube.show(savefile= timedir+"%s.pdf"%(cubefile.split('/')[-1].replace(".fits.gz",".pdf").replace(".fits",".pdf")))




#################################
#
#   MAIN 
#
#################################
if  __name__ == "__main__":
    print("see pysedm/bin/ccd_to_cube.py")



#! /usr/bin/env python
# -*- coding: utf-8 -*-


""" very hight level module that handles spaxel extraction target/host and calibration """

import numpy as np
from astropy import coordinates, units

from . import astrometry, io, sedm
class CubeParser():

    def __init__(self, filename):
        """ """
        self.load(filename)
    # ============== #
    #  Method        #
    # ============== #
    # --------- #
    #  LOADER   #
    # --------- #
    def load(self, filename):
        """ """
        self._filename = filename
        self._cube = sedm.load_sedmcube(filename)
        self._target_radec = coordinates.SkyCoord(self.cube.header["OBJRA"],
                                                  self.cube.header["OBJDEC"],
                                                  unit=(units.hourangle, units.deg))
        self._wcsifu = astrometry.WCSIFU.from_filename(filename)

    def load_photoifu(self, mag=18):
        """ """
        from photoifu import hostparser
        self._photoifu = hostparser.HostParser([self._target_radec.ra.deg, self._target_radec.dec.deg], mag, self.cube.filename)
        self._load_subcubes_()
        
    def load_fluxcalibration(self, fluxcalfile=None):
        """ """
        if fluxcalfile is None:
            self._fluxcalfile = io.fetch_nearest_fluxcal(file = self.cube.filename)

        from .fluxcalibration import load_fluxcal_spectrum
        self._fluxcal = load_fluxcal_spectrum( self._fluxcalfile )

    def _load_subcubes_(self, buffer=8):
        """ """
        self._subcubes = {"host": self.cube.get_partial_cube( self.photoifu.ifuproject.get_spaxels_in("host"), np.arange( len((self.cube.lbda))) ),
                          "target": self.cube.get_partial_cube( self.photoifu.ifuproject.get_target_pixels(), np.arange( len((self.cube.lbda))) )}
        
    # --------- #
    #  GETTER   #
    # --------- #
    def extract_spectra(self, nofluxcal=False, **kwargs):
        """ """
        from pyifu import get_spectrum
        
        self.cube_target.extract_pointsource(**kwargs)
        es_star = self.cube_target.extractstar
        
        cube_bkgd = es_star.es_products["bkgdmodel"]
        self.spec_sky_raw = cube_bkgd.get_spectrum(cube_bkgd.get_faintest_spaxels(1))
        self.spec_sky = sedm.flux_calibrate_sedm(self.spec_sky_raw, fluxcalfile=self._fluxcalfile, nofluxcal=nofluxcal)
        
        self.spec_target_raw = es_star.es_products["spec"]
        self.spec_target = sedm.flux_calibrate_sedm(self.spec_target_raw, fluxcalfile=self._fluxcalfile, nofluxcal=nofluxcal)
        
        mean_host = self.cube_host.get_spectrum(np.arange( self.cube_host.nspaxels ))
        self.spec_host_raw = get_spectrum(mean_host.lbda, mean_host.data- self.spec_sky.data,
                                            variance=mean_host.variance, header=mean_host.header)
        self.spec_host = sedm.flux_calibrate_sedm(self.spec_host_raw, fluxcalfile=self._fluxcalfile, nofluxcal=nofluxcal)
        
        

    # ============== #
    #  Properties    #
    # ============== #
    @property
    def cube(self):
        """ """
        if not hasattr(self,"_cube"):
            self._cube = None
        return self._cube
    
    @property
    def fluxcalibration(self):
        """ """
        if not hasattr(self,"_fluxcal"):
            self.load_fluxcalibration()
        return self._fluxcal

    @property
    def fluxcalfile(self):
        """ """
        if not hasattr(self,"_fluxcalfile"):
            self.load_fluxcalibration()
        return self._fluxcalfile

    @property
    def cube_target(self):
        """ """
        if not hasattr(self,"_subcubes"):
            return None
        return self._subcubes["target"]
    
    @property
    def cube_host(self):
        """ """
        if not hasattr(self,"_subcubes"):
            return None
        return self._subcubes["host"]
    
    @property
    def wcsifu(self):
        """ """
        if not hasattr(self,"_wcsifu"):
            self._wcsifu = None
        return self._wcsifu

    @property
    def photoifu(self):
        """ """
        if not hasattr(self,"_photoifu"):
            self.load_photoifu()
        return self._photoifu

    



""" Basic Dask Tools  """

import warnings
import pandas
import numpy as np
import dask

from .. import io, get_sedmcube
from .. import fluxcalibration
from ..io import REDUXPATH
import glob
import os
from astropy.io.fits import getval


def fetch_hypergalfluxcal(date, range_night=2):
    """ """
    from datetime import datetime, timedelta
    files = []
    d_ = datetime.strptime(date, '%Y%m%d')
    dates = [(d_ + timedelta(n)).strftime('%Y%m%d').replace('-', '')
             for n in range(-range_night, range_night+1)]
    for date_ in dates:
        for file in glob.glob(os.path.join(REDUXPATH, date_)+"/fluxcal_hypergal*"):
            files.append(file)
    return files


def fetch_pysedmfluxcal(date, range_night=0):
    """ """
    from datetime import datetime, timedelta
    files = []
    d_ = datetime.strptime(date, '%Y%m%d')
    dates = [(d_ + timedelta(n)).strftime('%Y%m%d').replace('-', '')
             for n in range(-range_night, range_night+1)]
    
    for date_ in dates:
        for file in glob.glob(os.path.join(REDUXPATH, date_)+"/fluxcal*"):
            files.append(file)

    if len(files) == 0:
        warnings.warn(f"No fluxcal found for night {date}")
    return files


def get_fluxcal_file(cube, hgfirst=False, update=False):
    """ """
    date = cube.header["OBSDATE"].replace("-", "")
    # - Make sure you have all the fluxcalibration files locally
    if hgfirst:  # Use hypergal fluxcalibration file if available
        filefluxcal = fetch_hypergalfluxcal(date)
        for f_ in filefluxcal:
            try:
                spec = fluxcalibration.load_fluxcal_spectrum(f_)
                if np.min(spec.data) < 0:
                    filefluxcal.remove(f_)
            except TypeError:
                filefluxcal.remove(f_)
        if len(filefluxcal) == 1:
            return filefluxcal[0]
        elif len(filefluxcal) > 1:
            fluxcal_mjd_obs = [getval(f, "MJD_OBS") for f in filefluxcal]
            mjd = cube.header.get("MJD_OBS")
            target_mjd_obs = mjd
            return filefluxcal[np.argmin(np.abs(target_mjd_obs - np.asarray(fluxcal_mjd_obs)))]
    from ztfquery import sedm
    squery = sedm.SEDMQuery()
    # downloads what is missing.
    try:
        import json
        _ = squery.get_night_fluxcal(date, nprocess=1, show_progress=False)
    except (json.JSONDecodeError, OSError) as e:
        warnings.warn(f"Corrupted file from whatdata at date {date}")
    # - grab the nearest
    try:
        fluxcal = io.fetch_nearest_fluxcal(mjd=cube.header.get("MJD_OBS"))
        
    except:
        print("enter inside the except.")
        filefluxcal = fetch_pysedmfluxcal(date)
        for f_ in filefluxcal:
            try:
                spec = fluxcalibration.load_fluxcal_spectrum(f_)
                if np.min(spec.data) < 0:
                    filefluxcal.remove(f_)
            except TypeError:
                filefluxcal.remove(f_)
            except FileNotFoundError:
                filefluxcal.remove(f_)
                warnings.warn(f"{f_} is a Corrupted file from whatdata at date {date}")

        if len(filefluxcal) == 0:
            raise ValueError("no filefluxcal found.")
        
        if len(filefluxcal) == 1:
            return filefluxcal[0]
        
        elif len(filefluxcal) > 1:
            fluxcal_mjd_obs = [getval(f, "MJD_OBS") for f in filefluxcal]
            mjd = cube.header.get("MJD_OBS")
            target_mjd_obs = mjd
            fluxcal = filefluxcal[np.argmin(
                np.abs(target_mjd_obs - np.asarray(fluxcal_mjd_obs)))]
            
    return fluxcal


def calibrate_cube(cube, fluxcalfile, airmass=None, backup_airmass=1.1, store_data=False):
    """ """
    if airmass is None:
        airmass = cube.header.get("AIRMASS", backup_airmass)

    fluxcal = fluxcalibration.load_fluxcal_spectrum(fluxcalfile)
    cube.scale_by(fluxcal.get_inversed_sensitivity(cube.header.get("AIRMASS", backup_airmass)),
                  onraw=False)
    cube.set_filename(cube.filename.replace("e3d", "cale3d"))
    header = {**dict(cube.header), **
              dict({'FLUXCAL': f'{os.path.basename(fluxcalfile)}'})}
    cube.set_header(header)

    if store_data:
        cube.writeto(cube.filename)

    return cube


def _no_delayed_(func):
    return func


def get_delayed_func(as_dask):
    """ """
    return dask.delayed if as_dask else _no_delayed_


class ClientHolder(object):

    def __init__(self, client=None):
        """ """
        if client is not None:
            self.set_client(client)

    def set_client(self, client):
        """ """
        self._client = client

    # =============== #
    #  Properties     #
    # =============== #
    @property
    def client(self):
        """ Dask client used for computation """
        if not hasattr(self, "_client"):
            return None
        return self._client

#
#
#
#
#


class _SEDMFileHolder_(ClientHolder):

    def __init__(self, client=None, files=None):
        """ """
        _ = super().__init__(client=client)
        if files is not None:
            self.set_files(files)
    # =============== #
    # Initialisation  #
    # =============== #

    @classmethod
    def from_files(cls, files, client):
        """ shortcut to __init__ for simplicity """
        return cls(client=client, files=files)

    @classmethod
    def from_name(cls, name, client, **kwargs):
        """ """
        files = cls.get_files(client=client, targetname=name, **kwargs)
        return cls.from_files(files=files, client=client)

    @classmethod
    def from_date(cls, date, client, **kwargs):
        """ """
        files = cls.get_files(client=client, dates=date, **kwargs)
        return cls.from_files(files=files, client=client)

    @classmethod
    def from_daterange(cls, daterange, client, **kwargs):
        """ """
        from ztfquery import sedm
        start, end = daterange
        dates = sedm.build_datelist(start=start, end=end)
        return cls.from_date(dates, client, **kwargs)

    @classmethod
    def from_month(cls, year, month, client, **kwargs):
        """ """
        from calendar import monthrange
        monthdates = [f'{year:04d}{month:02d}{d:02d}' for d in range(
            1, monthrange(year, month)[1] + 1)]
        return cls.from_date(monthdates, client=client, **kwargs)

    @classmethod
    def from_year(cls, year, client, **kwargs):
        """ """
        return cls.from_daterange([f"{year}-01-01", f"{year}-12-31"], client, **kwargs)

    # - Static Get Files
    @staticmethod
    def get_files(client, dates=None, targetname=None,  **kwargs):
        """ """
        raise NotImplementedError("get_file mush be implemented")

    # =============== #
    # Methods         #
    # =============== #
    # -------- #
    #  SETTER  #
    # -------- #
    def set_files(self, files, unique=True):
        """ """
        if len(files) == 0:
            raise ValueError("no files given")

        # cubefiles
        self._files = files if not unique else list(np.unique(files))
        # dataziles
        datafile = pandas.DataFrame(self.files, columns=["filepath"])
        dataall = datafile["filepath"].str.split("/", expand=True)
        datafile["basename"] = dataall[dataall.columns[-1]].values

        info = pandas.DataFrame.from_records(
            datafile["basename"].apply(io.parse_filename))
        datafile = pandas.merge(
            datafile, info, left_index=True, right_index=True)
        datafile["is_std"] = datafile["name"].str.contains("STD")
        self._datafiles = datafile

    # =============== #
    #  Properties     #
    # =============== #
    @property
    def files(self):
        """ """
        if not hasattr(self, "_files"):
            return None
        return self._files

    def has_files(self):
        """ """
        return self.files is not None and len(self.files) > 0

    @property
    def datafiles(self):
        """ stored version of get_datafiles """
        if not hasattr(self, "_datafiles"):
            return None
        return self._datafiles


class DaskCCD(_SEDMFileHolder_):
    """ """
    # =============== #
    # Initialisation  #
    # =============== #
    @staticmethod
    def get_files(client, dates=None, targetname=None,  **kwargs):
        """ """
        if targetname is None and dates is None:
            raise ValueError("either dates or targetname must be given")

        from ztfquery import sedm
        squery = sedm.SEDMQuery()
        files = []
        if dates is not None:
            files += list(squery.get_night_crr(dates, client=client, **kwargs))

        if targetname is not None:
            files += list(squery.get_target_crr(targetname, client=client))

        return files


class DaskCube(_SEDMFileHolder_):

    # =============== #
    # Initialisation  #
    # =============== #

    @staticmethod
    def get_files(client, dates=None, targetname=None, incl_astrom=True, **kwargs):
        """ """
        if targetname is None and dates is None:
            raise ValueError("either dates or targetname must be given")

        from ztfquery import sedm
        squery = sedm.SEDMQuery()
        cubes = []
        astrom = []
        if dates is not None:
            cubes += list(squery.get_night_cubes(dates,
                          client=client, **kwargs))
            if incl_astrom:
                astrom += list(squery.get_night_astrom(dates, client=client))
        if targetname is not None:
            cubes += list(squery.get_target_cubes(targetname, client=client))
            if incl_astrom:
                astrom += list(squery.get_target_astrom(targetname,
                               client=client))

        return cubes

    # -------- #
    #  SETTER  #
    # -------- #

    # -------- #
    #  GETTER  #
    # -------- #
    def get_std_basename(self, excluse_std=None, avoid_noncalspec=True, avoid_bad=True):
        """ get the cubefile_dataframe entry associated to the standard stars """
        dstd = self.datafiles[self.datafiles["is_std"]]

        if excluse_std is not None:
            re_exclude = "|".join(list(np.atleast_1d(excluse_std)))
            dstd = dstd[~dstd["name"].str.contains(re_exclude)]

        if avoid_noncalspec:
            noncalspec_std = io.get_noncalspec_standards()
            dstd = dstd[~dstd["name"].str.contains("|".join(noncalspec_std))]

        if avoid_bad:
            bad_std = io.get_bad_standard_exposures()
            dstd = dstd[~dstd["basename"].str.contains("|".join(bad_std))]

        return dstd["basename"]

    def get_datafiles(self, excluse_std=None,
                      avoid_noncalspec=True, avoid_bad=True,
                      add_stdcalib=True, index_per_calib=True):
        """ """
        datafiles = self.datafiles.copy()
        if add_stdcalib:
            std_datafiles = self.get_std_basename(excluse_std=excluse_std,
                                                  avoid_noncalspec=avoid_noncalspec,
                                                  avoid_bad=avoid_bad)
            df_std = datafiles.loc[std_datafiles.index]
            id_ = np.argmin(
                np.abs(datafiles["mjd"].values-df_std["mjd"].values[:, None]), axis=0)
            datafiles["std_calib"] = df_std["basename"].iloc[id_].values

            if index_per_calib:
                return datafiles.set_index(["std_calib", datafiles.index])

        return datafiles

    # -------- #
    #  DASK    #
    # -------- #
    @staticmethod
    def get_calibrated_cube(cubefile_, fluxcalfile=None, hgfirst=False,
                            apply_byecr=True, get_filename=False, as_dask=True, **kwargs):
        """ """
        # dask.delayed or doing nothing
        delayed = get_delayed_func(as_dask)

        # 1. Get cube
        cube = delayed(get_sedmcube)(cubefile_, apply_byecr=apply_byecr)

        # 2. Get flux calibration file (if any)
        if fluxcalfile is None:
            fluxcalfile = delayed(get_fluxcal_file)(
                cube, hgfirst)  # could be None

        # 3. Flux calibrating the cube
        calibrated_cube = delayed(calibrate_cube)(cube, fluxcalfile, **kwargs)

        if get_filename:
            return calibrated_cube.filename

        return calibrated_cube



""" Dasked version of pysedm/bin/extractstars.py """

import os
import numpy as np
from dask import delayed
from .base import DaskCube


from .. import get_sedmcube, fluxcalibration, io
from ..sedm import SEDMExtractStar, flux_calibrate_sedm
from astropy.io import fits


def get_extractstar(cube, step1range=[4500,7000], step1bins=6,
                   centroid="auto", **kwargs):
    """ """
    es = SEDMExtractStar(cube)
    es.set_lbdastep1(lbdarange=step1range, bins=step1bins)
    es.set_centroid(centroid=centroid)
    return es

def get_fluxcalfile(cubefile):
    """ """
    return io.fetch_nearest_fluxcal(mjd=fits.getval(cubefile,"MJD_OBS"))

def run_extractstar(es, spaxelbuffer = 10,
                    spaxels_to_use=None,
                    spaxels_to_avoid=None,
                    psfmodel="NormalMoffatTilted",
                    slice_width = 1, fwhm_guess=None, 
                    verbose=False,
                    store_fig=True, store_data=True,
                    **kwargs):
    """ """
    
    if spaxels_to_use is not None: # You fixed which you want
        if len(spaxels_to_use)<4:
            warnings.warn("ExtractStar: You provided less than 4 spaxel to fit")
        es.set_fitted_spaxels(spaxels_to_use)

    elif es.fitted_spaxels is None: # Automatic selections (with or without spaxels)
        es.get_spaxels_tofit(buffer=spaxelbuffer, update=True,
                             spaxels_to_avoid=spaxels_to_avoid)

    if verbose: print("* Starting extractstar.run")
    es.run(slice_width=slice_width, psfmodel=psfmodel,
           fwhm_guess=fwhm_guess, verbose=verbose, **kwargs)
    
    rawspec = es.get_spectrum("raw", persecond=True)
    rawspec.set_filename(es.cube.filename.replace("e3d", "esspec"))
    #
    # - Output
    if store_fig and es is not None:
        basename = es.basename
        es.show_mla(savefile = basename.replace("{placeholder}","esmla")+".pdf")
        if es._es_headerkey["posok"]:
            es.show_adr(savefile = basename.replace("{placeholder}","esadr")+".pdf")
            es.show_psf(savefile = basename.replace("{placeholder}","espsf")+".pdf")
        rawspec.show(savefile = basename.replace("{placeholder}","esspec")+".pdf")
        
    if store_data:
        rawspec.writeto(rawspec.filename)

    return rawspec

def calibrate_spec(spec, fluxcalfile, default_airmass=1.1,
                       store_fig=True, store_data=True):
    """ """
    if fluxcalfile is None:
        fluxcal = None
    else:
        fluxcal = fluxcalibration.load_fluxcal_spectrum(fluxcalfile)
        
    if fluxcal is None:
        spec.header["FLUXCAL"] = (False,"has the spectra been flux calibrated")
        spec.header["CALSRC"] = (None, "Flux calibrator filename")
        spec.header["BUNIT"]  = (spec.header.get('BUNIT',""),"Flux Units")
        specout = spec.copy()
    else:
        spec.scale_by(fluxcal.get_inversed_sensitivity(spec.header.get("AIRMASS",default_airmass)),
                      onraw=False)
        spec.header["FLUXCAL"] = (True, "has the spectra been flux calibrated")
        spec.header["CALSRC"] = (os.path.basename(fluxcalfile), "Flux calibrator filename")
        spec.header["BUNIT"]  = ("erg/s/A/cm^2", "Flux Units")
        specout = spec.copy()

    specout.set_filename( spec.filename.replace("esspec","spec") )
    if store_fig:
        specout.show( savefile=specout.filename.replace(".fits",".pdf") )
                         
    if store_data:
        specout.writeto(specout.filename)
        
    return specout
        
def build_fluxcalibrator(rawspec_std,
                             store_plot=True, store_data=True):
    """ 
    

    Returns
    -------
    speccal filename
    """
    speccal, fl = fluxcalibration.get_fluxcalibrator(rawspec_std, fullout=True)
    speccal.header.set("SOURCE", os.path.basename(rawspec_std.filename),
                           "This object has been derived from this file")
    speccal.header.set("PYSEDMT","Flux Calibration Spectrum", "Object to use to flux calibrate")

    speccal.set_filename( rawspec_std.filename.replace("esspec","fluxcal") )
    if store_plot:
        _ = fl.show( savefile=speccal.filename.replace(".fits",".pdf") )
        
    if store_data:
        speccal.writeto( speccal.filename )

    # Returning the filename because dask has serialization issue 
    return speccal.filename



class DaskES( DaskCube ):
    """
    1. Get Files
    cubefiles

    2. Get Cubes
    cube = get_sedmcube(filecube)
    cube.header['REDUCER'] = args.reducer
    
    3. remove cubes for CR

    4. Extract point source from the cubes 
    es_out = cube.extract_pointsource(**es_options)
    
    5. Get flux calibration file

    6. flux calibrate the results

    7. store the result
    """


    # =============== #
    #    Methods      #
    # =============== #


    # -------- #
    # Running  #
    # -------- #
    def compute(self, stdbasenames=None, get_delayed=False, **kwargs):
        """ """
        if stdbasenames is None:
            data = self.get_datafiles(add_stdcalib=True, index_per_calib=True, **kwargs)
            stdbasenames = np.unique(data.index.levels[0])
        
        spectra = [self.stdconnected_extractstars(std_basename=stdbasename)
                    for stdbasename in stdbasenames]
        
        if get_delayed:
            return spectra
        
        return self.client.compute(spectra)

    def stdconnected_extractstars(self, std_basename=None, **kwargs):
        """ """
        #
        # Get the cube paths
        datafile_std  = self.get_datafiles(**kwargs).xs(std_basename).set_index("basename")
        
        cubefile_std = datafile_std.loc[std_basename]["filepath"]
        cubefiles    = datafile_std.drop(std_basename)["filepath"].values

        #
        # Build the graph
        raw_std = self.extract_rawstar(cubefile_std, centroid="brightest")
        fluxcalfile = self.build_fluxcalibrator(raw_std)
        f_spec = [raw_std]
        for cubefile in cubefiles:
            rawspec = self.extract_rawstar(cubefile)
            spec = self.calibrate_rawspec(rawspec, fluxcalfile=fluxcalfile)
            f_spec.append(spec)
            
        return f_spec
    
    @classmethod
    def single_extractstars(cls, cube_filename, fluxcalfile=None, calibrate=True, **kwargs):
        """ """
        rawspec = cls.extract_rawstar(cube_filename, **kwargs)
        if fluxcalfile is None and calibrate:
            fluxcalfile = delayed(get_fluxcalfile)(cube_filename)
            
        return cls.calibrate_rawspec(rawspec, fluxcalfile)

    # -------- #
    # Static   #
    # -------- #
    @staticmethod
    def extract_rawstar(cube_filename, centroid='auto', store_fig=True, store_data=True, **kwargs):
        """ Extract the pointsource from the given filename 
        
        Parameters
        ----------
        cube_filename: [sting]
            Path to the e3d cube.

        Returns
        -------
        Spectrum (not flux calibrated)
        """
        # 1. Get cube
        cube = delayed(get_sedmcube)(cube_filename)
                
        # 3. Get ExtractStar
        estar = delayed(get_extractstar)(cube, centroid=centroid)
        
        # 4. Run ExtractStar
        rawspec = delayed(run_extractstar)(estar, store_fig=store_fig, store_data=store_data,
                                               **kwargs)
        return rawspec

    @staticmethod
    def calibrate_rawspec(rawspec, fluxcalfile):
        """ """
        return delayed(calibrate_spec)(rawspec, fluxcalfile)

    @staticmethod
    def build_fluxcalibrator(rawspec):
        """ """
        return delayed(build_fluxcalibrator)(rawspec)
        


    # -------- #
    # Checks   #
    # -------- #
    def compute_check(self, futures):
        """ """
        



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
This script is to remove cosmic ray spaxels in the SEDM datacube.
-v20210204: 1st version.
"""

#################################
#
#   MAIN
#
#################################
if  __name__ == "__main__":

    import argparse
    import numpy as np

    from pysedm import get_sedmcube, io, byecr

    # ================= #
    #   Options         #
    # ================= #
    parser = argparse.ArgumentParser(
        description=""" run the interactive plotting of a given cube""",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('infile', type=str, default="None",
                        help='cube filepath')

    parser.add_argument('--inputcube',   action="store_true", default=False,
                        help='Is the input a cube ?')

    parser.add_argument('--auto',  type=str, default=None,
                        help='Shall this run an automatic PSF extraction')

    # Cosmic ray removal with byecr.py
    parser.add_argument('--byecr_lbda', type=float, default=None,
                         help="To check cosmic rays at a given wavelength.")

    parser.add_argument('--byecr_cut', type=float, default=5.0,
                         help="Cut criteria for byecr module.")

    parser.add_argument('--byecr_wspectral', action="store_true", default=False,
                         help="Also use a spectral filtering for byecr.")

    parser.add_argument('--byecr_showcube', action="store_true", default=False,
                         help="Show a cube with detected cosmic rays.")

    args = parser.parse_args()

    # ================= #
    #   The Scripts     #
    # ================= #

    # --------- #
    #  Date     #
    # --------- #
    if not args.inputcube:
        date = args.infile

    # ===================== #
    # Running byecr   #
    # ===================== #

    if args.auto is not None and len(args.auto) >0:

        for target in args.auto.split(","):

            filecubes = io.get_night_files(date, "cube.*", target=target.replace(".fits","")) if not args.inputcube else [target]
            print("cube file from which the cosmic-ray will be removed [auto]: "+ ", ".join(filecubes))

            # - loop over the file cube
            for filecube in filecubes:
                print("Automatic removing cosmic-rays of target %s, file: %s"%(target, filecube))

            #
            # DATA IN
            #
                cube = get_sedmcube(filecube)

                print(" Starting byecr cosmic ray removal ".center(50, "-"))
                targetid = io.filename_to_id(filecube)
                byecrclass = byecr.get_cr_spaxels_from_byecr(date, targetid)
                cr_df = byecrclass.get_cr_spaxel_info(lbda_index=args.byecr_lbda,
                                              cut_criteria=args.byecr_cut,
                                              wspectral=args.byecr_wspectral)

                cube.data[cr_df["cr_lbda_index"], cr_df["cr_spaxel_index"]] = np.nan

                cube.header.set("NCR", len(cr_df), "total number of detected cosmic-rays from byecr")
                cube.header.set("NCRSPX", len(np.unique(cr_df["cr_spaxel_index"])), "total number of cosmic-ray affected spaxels")

                print("Number of detected cosmic rays = %i" %len(cr_df))
                print("Number of cosmic ray affected spaxels = %i" %len(np.unique(cr_df["cr_spaxel_index"])))

            # -
            # - SAVING
            # -
                savefile = filecube.replace("e3d_", "e3d_crr_")
                print("savefile = %s" %savefile)
                cube.writeto(savefile=savefile)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module is to get target or host spaxels in SEDM cube data.
-v.20210209: add 'remove target spaxel' for default 'get_others_spaxels()', e.g., 20210204_ZTF21aagtqna.
-v.20201113: add 'convex_hull' option, such as geometry.Polygon.convex_hull, after investigating 20201111_03_36__55_ZTFacdqjeq.
-v.20200702: 'get_others_spaxels': fix 'spaxel_id' option, when 'forced_addcontsep_mag' works automatically.
-v.20200622: consider when no other spaxels in contsep mag, e.g. 20200620_ZTF20abffaxl.
-v.20200619: consider 'target_others_dist' for multiple ref sources.
-v.20200618: consider 'np.nanmin(target_others_dist) > 2' when 'refimg_coords_in_ifu.size' == 0.
-v.20200612: From 20200611_ZTF18aaiykoz, which show mostly overlappbed coordinates b/w the target and ref sources.
             1. add criteria such that 'np.nanmin(target_others_dist) > 2'. '2' is from minimum size of polygon diameter.
             > for that case, "self.target_contsep_mag = -99.0".
-v.20200609: From 20200609_ZTF20abahhml, which has 'nan' in contour arrays,
             1. add "not np.isnan(t_).any()" and change "geometry.Polygon(t_).area>2 (before 1)"
                for '_ifucounts_cleaned'.
-v.20200608: From 20200604_ZTF20abaszgh, which has lots of segemented small polygons,
             1. add '_ifucounts_cleaned'.
             2. For counting method, add 'if poly_.area < 100:'.
-v.20200427-28: 1.show_ifudata: add showing 'offset', 'contsep_mag' and 'fake_mag'.
             2. add 'fake_mag' (affects target) and 'forced_mag' (affects host) option.
-v.20200420: 'offset' option.
-v.20200407.
"""

import numpy as np
import matplotlib.pyplot as mpl
from shapely import geometry

from . import io
from . import astrometry
from . import sedm

def get_spaxels_from_constsep(date, targetid, offset=None, fake_mag=None, forced_mag=None, isomag=None):
    """
    Parameters
    ----------
    date: str
        YYYYMMDD
    targetid: str
        unique part of the filename, i.e. ZTF name or timestamp like 06_10_45
    offset: list of 2 floats, default=[0.0, 0.0],
        x and y offset between the cube and reference images.
    fake_mag: float, default=16.0,
        fake target mag for the reference image.
    forced_mag: float, default=0.0, (1.0 is recommended to use host spectra)
        add forced_mag to contsep_mag in steps of 0.5 to get more host spaxels.
        ! if contpse_mag = 26 mag (the faintest), it raises error.

    Return
    ----------
    SEDM_CONTOUR class
    """

    if offset is None:
        offset = [0.0, 0.0]
    if fake_mag is None:
        fake_mag = 16.0
    if forced_mag is None:
        forced_mag = 0.0
    if isomag is None:
        isomag = [20, 26, 13]

    return SEDM_CONTOUR.from_sedmid(date, targetid, offset, fake_mag, forced_mag, isomag)

####################
#                  #
#   CLASS          #
#                  #
####################

class SEDM_CONTOUR():

    #
    # Initializing
    #

    def __init__(self, astromfile, cube, offset, fake_mag, forced_mag, isomag_range):
        """
        "astromfile" will be provided with its full path.
        "isomag_range": lists to make an isomag contour, [start, end, step number]

        1. Set astrometry and download PS BytesIO data, so it takes time.
        2. Then, make a fake target in the reference PS image with 'fake_mag',
        3. Then, set isomag contour in reference PS data ("refcount") and IFU data ("ifucount").

        return self.target_consep_mag, self.target_polygon_loc

        """
        self.isomag = np.linspace(isomag_range[0], isomag_range[1], isomag_range[2])
        self.cube = cube

        self.set_target_astrometry(astromfile)
        self.set_ps_reference()
        self.set_offset(offset)

        self.set_fake_mag(fake_mag)
        self.set_fake_target_in_ref(self.fake_mag)
        self.set_ref_iso_contours()

        #self.set_forced_addcontsep_mag(forced_mag)
        self.forced_addcontsep_mag = forced_mag # to tweak host (others) contour.

    @classmethod
    def from_sedmid(cls, date, targetid, offset, fake_mag, forced_mag, isomag):
        """
        Parameters
        ----------
        date: str
            YYYYMMDD
        targetid: str
            unique part of the filename, i.e. ZTF name or timestamp like 06_10_45

        Returns
        -------
        Instance of the class (like a __init__)
        """
        filename = io.get_night_files(date, "cube", targetid)[0]
        cube = sedm.get_sedmcube(filename)

        return cls(filename, cube, offset, fake_mag, forced_mag, isomag)

    #
    #  SETTER
    #

    def set_target_astrometry(self, astromfile):
        """ """
        self.astro = astrometry.Astrometry(astromfile)

    def set_target_coords_in_ifu(self):
        """ """
        self._target_coords_in_ifu = self.iref.get_target_coordinate(where="ifu")

    def set_ps_reference(self):
        """ download PS BytesIO data (it takes time.). """
        self.iref = astrometry.IFUReference.from_astrometry(self.astro)

    def set_refimg_coords_in_ifu(self):
        """ """
        self._refimg_coords_in_ifu = self.iref.get_refimage_sources_coordinates("ifu", inifu=True)

    def set_target_others_dist(self):
        self._target_others_dist = []

        if self.refimg_coords_in_ifu.size > 0:
            for i in range(0, len(self.refimg_coords_in_ifu[0])):
                __target_others_dist = np.sqrt((self.target_coords_in_ifu[0] - self.refimg_coords_in_ifu[: ,i][0])**2 +
                                               (self.target_coords_in_ifu[1] - self.refimg_coords_in_ifu[: ,i][1])**2 )
                self._target_others_dist.append(__target_others_dist)
        else:
            # if there is no others in the cube, return half of cube size.
            self._target_others_dist = 20

    def set_offset(self, offset):
        """ set astrometry offset between reference image and cube image. """
        self.iref.set_ifu_offset( offset[0], offset[1] )
        self.offset = np.asarray( offset )

    def set_fake_mag(self, fake_mag):
        """ """
        self.fake_mag = fake_mag

    def set_fake_target_in_ref(self, fake_mag):
        """ get a fake target with magnitude of "fake_mag" in the reference PS imaga to make a contour for IFU data. """
        self.iref.photoref.build_pointsource_image(fake_mag)

    def set_ref_iso_contours(self):
        """ get iso mag contours in reference PS data ("refcounts"). """

        self.refcounts_ = self.iref.get_iso_contours(where="ref", isomag=self.isomag )

    def set_ifu_iso_contours(self):
        """ get iso mag contours in IFU data ("ifucounts"). """
        _ifucounts = self.iref.get_iso_contours(where="ifu", isomag=self.isomag )

        self._ifucounts_cleaned = {}
        for k, v in _ifucounts.items():
            self._ifucounts_cleaned[k] = [t_ for t_ in v if (len(t_)>4) and (not np.isnan(t_).any()) and (geometry.Polygon(t_).area>2) ]

    def set_target_contour_location(self):
        """
        find a location of a target contour in each isomag contour in the cube.

        Returns a list of [ (isomag, ifucount key of isomag, isomag index, 1 th array in isomag contour)]
        """

        __target_polygon_loc = []

        for mag_index in range( 0, len(self.ifucounts) ):

            for i in range(0, len(self.ifucounts[ list(self.ifucounts.keys())[mag_index] ])):
                poly = geometry.Polygon(self.ifucounts[ list(self.ifucounts.keys())[mag_index] ][i])
                target_polygon_ = poly.contains( geometry.Point( self.iref.get_target_coordinate("ifu")) )

                if target_polygon_:
                    __target_polygon_loc.append( (self.isomag[mag_index], list(self.ifucounts.keys())[mag_index], mag_index, i) )

        self._target_polygon_loc = np.array(__target_polygon_loc)

    #
    # RESET
    #

    def _reset_forced_addcontsep_mag(self, forced_mag):
        """ for host spaxels """
        self.forced_addcontsep_mag = forced_mag

    def _reset_fake_mag(self, fake_mag):
        """ """
        self.fake_mag = fake_mag
        self.set_fake_target_in_ref(self.fake_mag)
        self.set_ref_iso_contours()
        self.set_ifu_iso_contours()
        self.set_target_contour_location()

    #
    #  GETTER
    #

    def get_refimage_coords_in_ifu(self):
        """ get refimage coordinates in ifu, considering offset! """

        refimage_coords_in_ifu = self.iref.get_refimage_sources_coordinates("ifu", inifu=True)

        return refimage_coords_in_ifu

    def get_target_faintest_contour_w_area_method(self):
        """
        get a faintest contour for the target by comparing contour area
        ,between the i th and (i-1) th contours, starting from the faintest contour.

        Basic concept is that
        1) if areas for both contours are big (here >50), that means both contours include (most likely) with other sources, like a galxy.
        2) or if the ith contour area is as small as 10, that contour is for the target.
        3) or if the contour area difference (i th - (i-1) th) is larger than 40, i th contour is for the target.
        """

        mag_step_diff = len(self.ifucounts) - len(self.target_polygon_loc)

        for i in range(1, len(self.target_polygon_loc)+mag_step_diff):
            poly_ref = geometry.Polygon( self.ifucounts[self.target_polygon_loc[len(self.target_polygon_loc) - i ][1] ][int(self.target_polygon_loc[len(self.target_polygon_loc)- i ][3])] )

            i_comp = i + 1
            poly_comp = geometry.Polygon( self.ifucounts[ self.target_polygon_loc[len(self.target_polygon_loc) - i_comp ][1] ][int(self.target_polygon_loc[len(self.target_polygon_loc) - i_comp ][3])] )

            if (poly_ref.area>50) & (poly_comp.area>50):
                continue

            elif poly_ref.area<10:
                return self.target_polygon_loc[ len(self.target_polygon_loc) - i + mag_step_diff ][0]

            else:
                poly_criteria = np.abs(poly_ref.area - poly_comp.area)
                if poly_criteria > 40:
                    return self.target_polygon_loc[ len(self.target_polygon_loc) - i_comp + mag_step_diff ][0]

    def get_target_faintest_contour_w_counting_method(self):
        """
        get a faintest contour for the target by counting how many sources are in i th contour.
        We want to have only '1 target source' in i th contour.
        This is possible
        1) because we know the location of the target contour from the function "get_target_contour_location()"
        2) BUT in reference PS image, there is no target.
        """

        mag_step_diff = len(self.ifucounts) - len(self.target_polygon_loc)
        target_faintest_isomag_key = self.target_polygon_loc[len(self.target_polygon_loc)-1][1]

        #refimg_coords_in_ifu = self.get_refimage_coords_in_ifu()

        refimg_x = self.refimg_coords_in_ifu[0]
        refimg_y = self.refimg_coords_in_ifu[1]

        # starts from the faintest in target_polygon_loc
        for i in range(1, len(self.target_polygon_loc)+mag_step_diff):
            poly_ = geometry.Polygon( self.ifucounts[ self.target_polygon_loc[len(self.target_polygon_loc)-i][1] ][int(self.target_polygon_loc[len(self.target_polygon_loc)-i][3])] )

            if poly_.area < 100:
                chk_poly_contain_src = []
                for src_num in range(0, len(refimg_x)):
                    chk_poly_contain_src.append(poly_.contains( geometry.Point([refimg_x[src_num], refimg_y[src_num]]) ))

                if any(chk_poly_contain_src) is False:
                    return self.target_polygon_loc[len(self.target_polygon_loc)-i][0]

    def get_target_faintest_contour(self, method="both"):
        """
        NOTE: When there is only 1 marked target in the cube,
              counting method always returns the faintest defined isomag, e.g., here 26.0 mag.
              This case, let's try only with 'area' method: "get_target_faintest_contour(method='area')."
        """

        if method == "both":
            area_method = self.get_target_faintest_contour_w_area_method()
            counting_method = self.get_target_faintest_contour_w_counting_method()

            if area_method == counting_method:
                target_consep_mag_ = area_method
                return target_consep_mag_

            else:
                #raise ValueError("Check the number of sources in the cube. If 1, use method='area'.")
                target_consep_mag_ = counting_method
                return target_consep_mag_

        elif method == "area":
            area_method = self.get_target_faintest_contour_w_area_method()
            counting_method = "Not used."

            target_contsep_mag_ = area_method

            return target_contsep_mag_

        elif method == "counting":
            counting_method = self.get_target_faintest_contour_w_counting_method()
            area_method = "Not used."

            target_consep_mag_ = counting_method

            return target_consep_mag_


    def get_target_contsep_information(self, forced_addcontsep=False):
        """
        get target faintes contour information, like faintes contour magnitude and its location in the target polygon.
        """

        if np.nanmin(self.target_others_dist) > 2.0:

            ## When there is only 1 marked target in the cube, use 'area' method.
            if self.refimg_coords_in_ifu.size == 0:
                self.target_contsep_mag = self.get_target_faintest_contour(method="area")
            else:
                self.target_contsep_mag  = self.get_target_faintest_contour()

            if forced_addcontsep:
                self.target_contsep_mag = self.target_contsep_mag + self.forced_addcontsep_mag

            target_contsep_mag_index_ = int( self.target_polygon_loc[self.target_polygon_loc[:,0] == self.target_contsep_mag][0][2] )
            target_contsep_array_index_  = int( self.target_polygon_loc[self.target_polygon_loc[:,0] == self.target_contsep_mag][0][3] )
            target_contsep_key_  = int( self.target_polygon_loc[self.target_polygon_loc[:,0] == self.target_contsep_mag][0][1] )

            return target_contsep_mag_index_, target_contsep_array_index_, target_contsep_key_

        else:
            self.target_contsep_mag = -99.0

            target_contsep_mag_index_ = np.nan
            target_contsep_array_index_  = np.nan
            target_contsep_key_ = np.nan

            return target_contsep_mag_index_, target_contsep_array_index_, target_contsep_key_

    #
    # Host
    #

    def set_host_coords_in_ifu(self):
        """ host identification and get its coordnates in pixel from PS wcs """

        if self.refimg_coords_in_ifu.size == 0: # if there is no other source in the cube
            #self._host_coords_in_ifu = self.iref.get_refimage_sources_coordinates(inifu=False, where='ifu')
            self._host_coords_in_ifu = ()

        elif self.refimg_coords_in_ifu.size == 2: # if there is only one other souce in the cube (size=2, x and y), that is the host
            self._host_coords_in_ifu = (float(self.refimg_coords_in_ifu[0]), float(self.refimg_coords_in_ifu[1]))

        else:
            host_coords_in_pspixel = self.iref.photoref.get_host_coords(inpixels=True)
            host_candi_coords_in_rcpixel = self.iref.get_refimage_sources_coordinates(inifu=True, where='ref')

            ps_rc_diff = []

            for i in range(0, len(host_candi_coords_in_rcpixel[0])):
                x_diff = host_coords_in_pspixel[0] - host_candi_coords_in_rcpixel[0][i]
                y_diff = host_coords_in_pspixel[1] - host_candi_coords_in_rcpixel[1][i]
                ps_rc_diff.append(np.sqrt(x_diff**2 + y_diff**2))

            host_index = np.argmin(ps_rc_diff)

            self._host_coords_in_ifu = (self.refimg_coords_in_ifu[0][host_index], self.refimg_coords_in_ifu[1][host_index])

    def set_host_contour_location(self):
        """
        find a location of a host contour in each isomag contour in the cube.

        Returns a list of [ (isomag, ifucount key of isomag, isomag index, 1 th array in isomag contour)]
        """

        __host_polygon_loc = []

        for mag_index in range( 0, len(self.ifucounts) ):

            for i in range(0, len(self.ifucounts[ list(self.ifucounts.keys())[mag_index] ])):
                poly = geometry.Polygon(self.ifucounts[ list(self.ifucounts.keys())[mag_index] ][i])
                host_polygon_ = poly.contains( geometry.Point( self.host_coords_in_ifu ) )

                if host_polygon_:
                    __host_polygon_loc.append( (self.isomag[mag_index], list(self.ifucounts.keys())[mag_index], mag_index, i) )

        self._host_polygon_loc = np.array(__host_polygon_loc)


    def get_host_radec_in_degree(self):
        """
        ???
        """

    #
    # SHOW
    #

    def show_ref_ifu(self, offset=[0.0, 0.0], savefile=None):
        """
        !!! offset will be given by hand after a visual check !!!
        """

        if any(self.offset != offset) is True:
            self.iref.set_ifu_offset( offset[0], offset[1] )
            self.ifucounts
            self.target_coords_in_ifu
            self.refimg_coords_in_ifu
            self.host_coords_in_ifu
            self.offset = np.asarray( offset )
        self.iref.show()

        if savefile is not None:
            fig.savefig( savefile )

    def show_catdata(self, savefile=None):
        """ """

        fig = mpl.figure()

        ax = fig.add_subplot(111)

        ax.imshow(np.log10(self.iref.photoref.fakeimage), origin="lower", cmap=mpl.cm.binary)

        ii=0
        for i,couts_ in self.refcounts_.items():
            for n, contour in enumerate(couts_):
                ax.plot(contour[:, 0], contour[:, 1], linewidth=2, color="C%d"%ii)
            ii+=1

        ra,dec,mag = self.iref.photoref._pstarget.catdata[["raMean","decMean","rPSFMag"]].values.T
        x,y = self.iref.photoref.refimage.coords_to_pixel(ra,dec).T
        flagout = (mag<0)
        ax.scatter(x[~flagout],y[~flagout], marker=".", c=mag[~flagout])

        if savefile is not None:
            fig.savefig( savefile )

    def show_ifudata(self, wcontour=True, wtargetspaxel=False, wotherspaxel=False, whostspaxel=False, savefile=None):
        """ """

        self.get_target_contsep_information()

        fig = mpl.figure()

        ax = fig.add_subplot(111)
        _ = self.iref.show_slice(ax=ax, vmin="5", vmax="99")


        if wcontour:
            ii=0
            for couts_ in self.ifucounts.keys():
                for n, contour in enumerate(self.ifucounts[couts_]):
                    if n==0:
                        ax.plot(contour[:, 0], contour[:, 1], linewidth=2, color="C%d"%ii, zorder=8, label=self.isomag[ii])
                    else:
                        ax.plot(contour[:, 0], contour[:, 1], linewidth=2, color="C%d"%ii, zorder=8)
                ii+=1
            ax.legend()

        if wtargetspaxel:
            target_ids_from_contsep = self.get_target_spaxels(spaxels_id=True)
            spaxel_patches = self.cube._display_im_(axim=ax, vmin="5", vmax="99")

            for i in target_ids_from_contsep:
                spaxel_patches[i].set_edgecolor("red")
                spaxel_patches[i].set_linewidth(1)
                spaxel_patches[i].set_zorder(10)

        if wotherspaxel:
            others_ids_from_contsep = self.get_others_spaxels(spaxels_id=True)
            spaxel_patches = self.cube._display_im_(axim=ax, vmin="5", vmax="99")

            for i in others_ids_from_contsep:
                spaxel_patches[i].set_edgecolor("k")
                spaxel_patches[i].set_linewidth(1)
                spaxel_patches[i].set_zorder(9)

        if whostspaxel: # to be revised!!!
            # host identification
            if len(self.host_coords_in_ifu) > 0:
                ax.scatter(self.host_coords_in_ifu[0], self.host_coords_in_ifu[1],
                           color="pink", s=100, zorder=10)

            host_ids_from_contsep = self.get_host_spaxels(spaxels_id=True)
            spaxel_patches = self.cube._display_im_(axim=ax, vmin="5", vmax="99")

            for i in host_ids_from_contsep:
                spaxel_patches[i].set_edgecolor("pink")
                spaxel_patches[i].set_linewidth(1)
                spaxel_patches[i].set_zorder(9)

        if whostspaxel:
            t1 = "offset=(%.1f, %.1f), contsep_mag for target=%.1f mag (host=%.1f mag)" % (self.offset[0], self.offset[1], self.target_contsep_mag-self.forced_addcontsep_mag, self.host_contsep_mag)
            t2 = "fake_mag=%.1f mag, forced_mag=%.1f mag" % (self.fake_mag, self.forced_addcontsep_mag)
        else:
            t1 = "offset=(%.1f, %.1f), contsep_mag for target=%.1f mag (other=%.1f mag)" % (self.offset[0], self.offset[1], self.target_contsep_mag-self.forced_addcontsep_mag, self.target_contsep_mag)
            t2 = "fake_mag=%.1f mag, forced_mag=%.1f mag" % (self.fake_mag, self.forced_addcontsep_mag)

        ax.set_xlim(-20,19)
        ax.set_ylim(-24,22)

        ax.text(-20, 24.5, t1, fontsize=10)
        ax.text(-20, 22.5, t2, fontsize=10)

        if savefile is not None:
            fig.savefig( savefile )

    ######################
    # Method for Spaxels #
    ######################

    def get_target_spaxels(self, spaxels_id=True):
        """
        get target spaxels from the faintest consep.

        Parameters
        ----------
        spaxels_id: bool -optional-
            Deafult is return spaxel id (not index).
            If True, return spaxels idenx (for using pysedm).

        Returns
        -------
        Return spaxel ids or indexes in numpy array.
        """

        target_contsep_mag_index, target_contsep_array_index, target_contsep_key = self.get_target_contsep_information(forced_addcontsep=False)

        if np.nanmin(self.target_others_dist) > 2.0:
            target_contsep_poly = geometry.Polygon(self.ifucounts[ list(self.ifucounts.keys())[target_contsep_mag_index] ][target_contsep_array_index] )

        else:
            target_contsep_poly = geometry.Polygon(self.ifucounts[ list(self.ifucounts.keys())[0] ][0] )

        target_contsep_spaxel_index = self.cube.get_spaxels_within_polygon(target_contsep_poly.convex_hull) #spaxel index

        if spaxels_id:
            target_contsep_spaxel_ids = np.arange(self.cube.nspaxels)[np.isin(self.cube.indexes,target_contsep_spaxel_index)] #spaxel index to id for drawing.
            return target_contsep_spaxel_ids
        else:
            return np.array(target_contsep_spaxel_index)


    def get_others_spaxels(self, spaxels_id=True):
        """
        get other sources' spaxels (including host and so on).

        Parameters
        ----------
        spaxels_id: bool -optional-
            Deafult is return spaxel id (not index).
            If True, return spaxels idenx (for using pysedm).
        forced_addcontsep_mag: float in 0.5 step -optional-
            To change 'target_contsep_mag' to get more wider other sources spaxels.
            e.g., when SN is exploded very close to the bright center, the number of target spaxels are mostly less than 10.


        Returns
        -------
        Return spaxel ids or indexes in numpy array.
        """

        spaxels_id_for_forcedmag = spaxels_id

        if np.nanmin(self.target_others_dist) > 2.0:
            _target_contsep_spaxel_index = self.get_target_spaxels(spaxels_id=False)

            #if len(_target_contsep_spaxel_index) < 11: # when SN is exploded close to the bright center.

            if (self.forced_addcontsep_mag > 0.0) and (self.get_target_faintest_contour() < 26.0): # TO BE CAHGNED
                 target_contsep_mag_index, target_contsep_array_index, target_contsep_key = self.get_target_contsep_information(forced_addcontsep=True)
                 others_contsep_mag_index = target_contsep_mag_index
                 others_contsep_array_index = [i for i in range(0,len( self.ifucounts[ list(self.ifucounts.keys())[others_contsep_mag_index] ] )) ]

                 __others_contsep_spaxel_ids = []
                 _others_contsep_spaxel_index = []

                 for i in range(0, len(others_contsep_array_index)):
                     others_contsep_polys = geometry.Polygon(self.ifucounts[ list(self.ifucounts.keys())[others_contsep_mag_index] ][others_contsep_array_index[i]] )

                     __others_contsep_spaxel_index = self.cube.get_spaxels_within_polygon(others_contsep_polys.convex_hull) #spaxel index
                     if len(__others_contsep_spaxel_index) > 0:
                         _others_contsep_spaxel_index += __others_contsep_spaxel_index

                 others_contsep_spaxel_index = np.unique(_others_contsep_spaxel_index) #spaxel index

                # remove target spaxel
                 for spax_ in _target_contsep_spaxel_index:
                     others_contsep_spaxel_index = others_contsep_spaxel_index[others_contsep_spaxel_index != spax_]

            else: # default
                target_contsep_mag_index, target_contsep_array_index, target_contsep_key = self.get_target_contsep_information(forced_addcontsep=False)
                others_contsep_mag_index = target_contsep_mag_index
                others_contsep_array_index = [i for i in range(0,len( self.ifucounts[ list(self.ifucounts.keys())[others_contsep_mag_index] ] )) if i != target_contsep_array_index]

                __others_contsep_spaxel_ids = []
                _others_contsep_spaxel_index = []

                for i in range(0, len(others_contsep_array_index)):
                    others_contsep_polys = geometry.Polygon(self.ifucounts[ list(self.ifucounts.keys())[others_contsep_mag_index] ][others_contsep_array_index[i]] )

                    __others_contsep_spaxel_index = self.cube.get_spaxels_within_polygon(others_contsep_polys.convex_hull) #spaxel index
                    if len(__others_contsep_spaxel_index) > 0:
                        _others_contsep_spaxel_index += __others_contsep_spaxel_index

                others_contsep_spaxel_index = np.unique(_others_contsep_spaxel_index) #spaxel index

                # remove target spaxel
                for spax_ in _target_contsep_spaxel_index:
                    others_contsep_spaxel_index = others_contsep_spaxel_index[others_contsep_spaxel_index != spax_]

            if (others_contsep_spaxel_index.size < 10) and (self.forced_addcontsep_mag < 1.6):
                self.forced_addcontsep_mag += 0.5
                return self.get_others_spaxels(spaxels_id=spaxels_id_for_forcedmag)

            if spaxels_id:
                others_contsep_spaxel_ids = np.arange(self.cube.nspaxels)[np.isin(self.cube.indexes, others_contsep_spaxel_index)] #spaxel index to id for drawing.
                return others_contsep_spaxel_ids
            else:
                return others_contsep_spaxel_index

        else:
            self.target_contsep_mag = -99.0

            others_contsep_spaxel_index = np.empty(0)
            return others_contsep_spaxel_index


    def get_host_spaxels(self, spaxels_id=True, updated=False):
        """
        get host spaxels from the faintest consep (or + forced_addcontsep_mag).

        Parameters
        ----------
        spaxels_id: bool -optional-
            Deafult is return spaxel id (not index).
            If True, return spaxels idenx (for using pysedm).
        updated: If True, it means 'host_contsep_information' is updated, because of other sources in 'host_contsep_poly'.

        Returns
        -------
        Return host spaxel ids or indexes in numpy array.
        """

        _spaxels_id = spaxels_id

        if np.nanmin(self.target_others_dist) > 2.0:

            if (self.refimg_coords_in_ifu.size == 0) or (self.refimg_coords_in_ifu.size == 2):
                self.host_contsep_mag = self.target_contsep_mag
                return self.get_others_spaxels(spaxels_id=_spaxels_id)
            else:

            #if (self.forced_addcontsep_mag > 0.0) and (self.get_target_faintest_contour() < 26.0):

                if updated is False:
                    self._set_host_contsep_information(update=False)

                only_host_check = self._host_polygon_check(self.host_contsep_poly)

                if only_host_check != 1:
                    # means there are host+other src.
                    self._set_host_contsep_information(update=True)
                    self.get_host_spaxels(spaxels_id=_spaxels_id, updated=True)

                host_contsep_spaxel_index = self.cube.get_spaxels_within_polygon(self.host_contsep_poly) #spaxel index

                if spaxels_id:
                    host_contsep_spaxel_ids = np.arange(self.cube.nspaxels)[np.isin(self.cube.indexes, host_contsep_spaxel_index)] #spaxel index to id for drawing.
                    return host_contsep_spaxel_ids
                else:
                    return np.array(host_contsep_spaxel_index)

        else: # target_others_dist < 2.0
            self.host_contsep_mag = self.isomag[12]

            len_ = len(self.ifucounts[ list(self.ifucounts.keys())[12] ])
            array_len = []

            for i in range(0, len_):
                array_len.append(len(self.ifucounts[ list(self.ifucounts.keys())[12] ][i]))

            array_index = np.argmax(array_len)

            host_contsep_poly = geometry.Polygon(self.ifucounts[ list(self.ifucounts.keys())[12] ][array_index] )
            host_contsep_spaxel_index = self.cube.get_spaxels_within_polygon(host_contsep_poly.convex_hull) #spaxel index

            if spaxels_id:
                host_contsep_spaxel_ids = np.arange(self.cube.nspaxels)[np.isin(self.cube.indexes, host_contsep_spaxel_index)] #spaxel index to id for drawing.
                return host_contsep_spaxel_ids
            else:
                return np.array(host_contsep_spaxel_index)


    def _set_host_contsep_information(self, update=False):
        """
        set host contsep information, like host_contsep_mag, _index, _array_index,
            and 'host_contsep_poly'gon.

        Parameters
        ----------
        update: If True, 'host_contsep_poly' is re-selected with next 'host_contsep_mag_index', like applying forced_addcontsep_mag.
                It will raised if there are other sources in the 'host_contsep_poly'.

        """
        if update is False:
            target_contsep_mag_index, _, _ = self.get_target_contsep_information(forced_addcontsep=False)
            self.host_contsep_mag_index = target_contsep_mag_index
            self.host_contsep_array_index  = int( self.host_polygon_loc[self.host_polygon_loc[:,0] == self.target_contsep_mag][0][3] )
            self.host_contsep_mag = self.target_contsep_mag
        else:
            self.host_contsep_mag_index -= 1
            self.host_contsep_array_index = int( self.host_polygon_loc[self.host_polygon_loc[:,2] == self.host_contsep_mag_index][0][3] )
            self.host_contsep_mag = self.host_polygon_loc[self.host_polygon_loc[:,2] == self.host_contsep_mag_index][0][0]

        self.host_contsep_poly = geometry.Polygon(self.ifucounts[ list(self.ifucounts.keys())[self.host_contsep_mag_index] ][self.host_contsep_array_index] ).convex_hull

    def _host_polygon_check(self, host_contsep_poly):
        """
        Check there is only host in the host polygon.
        """

        refimg_x = self.refimg_coords_in_ifu[0]
        refimg_y = self.refimg_coords_in_ifu[1]

        chk_host_poly_contain_others = []

        for src_num in range(0, len(refimg_x)):
            chk_host_poly_contain_others.append(host_contsep_poly.contains( geometry.Point([refimg_x[src_num], refimg_y[src_num]]) ))

        return chk_host_poly_contain_others.count(True)




    ###################
    # Properties      #
    ###################

    @property
    def ifucounts(self):
        """
        IFU iso contours.
        If an offset has changed, re-set ifucounts considering the offset.
         """
        if (not hasattr(self, "_ifucounts_cleaned")) | any( (self.offset != self.iref.ifu_offset) ) :
            self.set_ifu_iso_contours()

        return self._ifucounts_cleaned

    @property
    def target_polygon_loc(self):
        """
        """
        if not hasattr(self, "_target_polygon_loc"):
            self.set_target_contour_location()

        return self._target_polygon_loc

    @property
    def target_coords_in_ifu(self):
        """
        """
        if ( not hasattr(self, "_target_coords_in_ifu") ) | any( (self.offset != self.iref.ifu_offset) ):
            self.set_target_coords_in_ifu()

        return self._target_coords_in_ifu

    @property
    def refimg_coords_in_ifu(self):
        """
        """
        if ( not hasattr(self, "_refimg_coords_in_ifu") ) | any( (self.offset != self.iref.ifu_offset) ):
            self.set_refimg_coords_in_ifu()

        return self._refimg_coords_in_ifu

    @property
    def target_others_dist(self):
        """
        """
        if not hasattr(self, "_target_others_dist"):
            self.set_target_others_dist()

        return self._target_others_dist

    @property
    def host_coords_in_ifu(self):
        """
        """
        if ( not hasattr(self, "_host_coords_in_ifu") ) | any( (self.offset != self.iref.ifu_offset) ):
            self.set_host_coords_in_ifu()

        return self._host_coords_in_ifu

    @property
    def host_polygon_loc(self):
        """
        """
        if not hasattr(self, "_host_polygon_loc"):
            self.set_host_contour_location()

        return self._host_polygon_loc



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pysedm
from pysedm       import get_sedmcube, io, fluxcalibration, sedm, astrometry
from pysedm.sedm  import IFU_SCALE_UNIT, SEDM_MLA_RADIUS
from pyifu import get_spectrum






def build_aperture_param(stored_picked_poly, default_bkgd_scale=1.4):
    """ Build an aperture spectroscopy parameters

    Parameters
    ----------
    stored_picked_poly: [list]
        list made by InteractivePlot:
        - [ [ "Circle",[[xcentroid, ycentroid], radius]],
            [ "Circle",[[xcentroid_i, ycentroid_i], radius_i]]
          ]
        It should contains 1 or 2 Circle entries:
        - If only 1, then the background will be defined by the annulus made of 1 and `default_bkgd_scale` times the picked circle.
        - If 2, the background will be the annulus defined by the difference between the 2 circles. 
          The centroid will be the average of the 2 picked circles. 
    
    default_bkgd_scale: [float] -optional-
        Default scale used to define the background if only 1 circle provided (see above)

    Returns
    -------
    2x2 numpy array 
        ([xcentroid, ycentroid], [aperture_radius, background_outer_radius])
        
    """
    idx_circles = [i for i,k in enumerate(stored_picked_poly) if k[0] == "Circle"]
    if len(idx_circles)>2:
        raise AttributeError("More than 2 Circle instance currently stored in plot_interactive. Cannot build a Aperture")
        
    elif len(idx_circles)<1:
        raise AttributeError("You need to have defined at least 1 circular aperture.")
    
    centroids = [stored_picked_poly[i][1][0] for i in idx_circles]
    radiuses  = [stored_picked_poly[i][1][1] for i in idx_circles]
    if len(idx_circles) == 2:
        return np.mean(centroids, axis=0), np.sort(radiuses)
    else:
        print("INFO: default background radius scale used (annulus made of [1,%s] times the aperture radius)"%default_bkgd_scale)
        return np.asarray(centroids[0]), radiuses[0]*np.asarray([1, default_bkgd_scale])





# ======================= #
#
#   Pipeline Steps        #
#
# ======================= #


# -------------------- #
#  Flux Calibration    #
# -------------------- #
def flux_calibrate(spec, fluxcalfile=None, nofluxcal=False):
    """ """
    if fluxcalfile in ["None"]:
        fluxcalfile = None

    if not nofluxcal:
        # Which Flux calibration file ?
        if fluxcalfile is None:
            from pysedm import io
            print("INFO: default nearest fluxcal file used")
            date = io.header_to_date(spec.header)
            fluxcalfile = io.fetch_nearest_fluxcal(date, spec.filename)
        else:
            print("INFO: given fluxcal used.")

        # Do I have a flux calibration file ?
        if fluxcalfile is None:
            print("ERROR: No fluxcal for night %s and no alternative fluxcalsource provided. Uncalibrated spectra saved."%date)
            spec.header["CALSRC"] = (None, "Flux calibrator filename")
            flux_calibrated=False
        else:
            from pysedm.fluxcalibration import load_fluxcal_spectrum
            fluxcal = load_fluxcal_spectrum( fluxcalfile ) 
            spec.scale_by( fluxcal.get_inversed_sensitivity(spec.header.get("AIRMASS", 1.1) ), onraw=False)
            spec.header["CALSRC"] = (fluxcal.filename.split("/")[-1], "Flux calibrator filename")
            flux_calibrated=True
            
    else:
        spec.header["FLUXCAL"] = (False,"has the spectra been flux calibrated")
        spec.header["CALSRC"] = (None, "Flux calibrator filename")
        flux_calibrated=False
        
    # Flux Calibration
    if flux_calibrated:
        spec.header["FLUXCAL"] = (True,"has the spectra been flux calibrated")
        spec.header["BUNIT"]  = ("erg/s/A/cm^2","Flux Units")
    else:
        spec.header["FLUXCAL"] = (False,"has the spectra been flux calibrated")
        spec.header["BUNIT"]  = (spec.header.get('BUNIT',""),"Flux Units")
    return spec, flux_calibrated

#################################
#
#   MAIN 
#
#################################
if  __name__ == "__main__":
    
    import argparse
    import numpy as np
    from shapely      import geometry
    #
    import psfcube
    from psfcube      import script
    
    # ================= #
    #   Options         #
    # ================= #
    parser = argparse.ArgumentParser(
        description=""" run the interactive plotting of a given cube""",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('infile', type=str, default="None",
                        help='cube filepath')

    # // AUTOMATIC EXTRACTION
    parser.add_argument('--inputcube',   action="store_true", default=False,
                        help='Is the input a cube ?')
    
    parser.add_argument('--reducer',  type=str, default=io.SEDM_REDUCER,
                        help='What is your name? This will be added in the header [default: $SEDM_REDUCER]')

    #  which extraction
    parser.add_argument('--aperture',  type=str, default=None,
                        help='Do you want a simple aperture extraction ?')
    parser.add_argument('--ap_bkgdscale',  type=float, default=1.4,
                        help='Default width scale of the outter background ring [default 1.4]')

    # Auto PSF
    parser.add_argument('--auto',  type=str, default=None,
                        help='Shall this run an automatic PSF extraction')

    parser.add_argument('--autorange',  type=str, default="4500,7000",
                        help='Wavelength range [in Angstrom] for measuring the metaslice PSF')
    
    parser.add_argument('--autobins',  type=int, default=7,
                        help='Number of bins within the wavelength range (see --autorange)')

    parser.add_argument('--buffer',  type=float, default=8,
                        help='Radius [in spaxels] of the aperture used for the PSF fit. (see --centroid for aperture center)')

    parser.add_argument('--psfmodel',  type=str, default="NormalMoffatTilted",
                        help='PSF model used for the PSF fit: NormalMoffat{Flat/Tilted/Curved}')
    
    parser.add_argument('--normed',  action="store_true", default=False,
                        help='[DEV USE ONLY]: apply psfcube normalization on the psf profile.')

    # Apperture


    # // Generic information
    parser.add_argument('--centroid',  type=str, default="None",nargs=2,
                        help='Where is the point source expected to be? using the "x,y" format. If None, it will be guessed.'+
                            "\nGuess works well for isolated sources.")
    
    parser.add_argument('--centroiderr',  type=str, default="None",nargs=2,
                        help='What error do you expect on your given centroid.'+
                            '\nIf not provided, it will be 3 3 for general cases and 5 5 for maximum brightnes backup plan')
    parser.add_argument('--maxpos', action="store_true", default=False,
                        help='Set to use brightest spaxel for position')
    
    parser.add_argument('--lstep',  type=int, default=1,
                        help='Slice width in lbda step: default is 1, use 2 for fainter source and maybe 3 for really faint target')
    
    parser.add_argument('--display',  action="store_true", default=False,
                        help='Select the area to fit using the display function.')

    parser.add_argument('--vmin',  type=str, default="2",
                        help='Data Percentage used for imshow "vmin" when using the --display mode')
    
    parser.add_argument('--vmax',  type=str, default="98",
                        help='Data Percentage used for imshow "vmax" when using the --display mode')
    
    parser.add_argument('--tag',  type=str, default="None",
                        help='Add the tag on output filename.')

    # - Standard Star object
    parser.add_argument('--std',  action="store_true", default=False,
                        help='Set this to True to tell the program you what to build a calibration spectrum from this object')
    
    parser.add_argument('--fluxcalsource',  type=str, default="None",
                        help='Path to a "fluxcal" .fits file. This file will be used for the flux calibration. If nothing given, the nearest (in time and within the night) fluxcal will be used.')
    

    parser.add_argument('--nofluxcal',  action="store_true", default=False,
                        help='No Flux calibration')
    
    parser.add_argument('--notroncation',  action="store_true", default=False,
                        help='Do not apply the blue and red end edge troncation on flux calibrated spectra')

    
    parser.add_argument('--nofig',    action="store_true", default=False,
                        help='')

        
    args = parser.parse_args()
        
    # ================= #
    #   The Scripts     #
    # ================= #

    # --------- #
    #  Date     #
    # --------- #
    if not args.inputcube:
        date = args.infile

    
    # ================= #
    #   Actions         #
    # ================= #
    extracted_objects = []
    # ---------- #
    # Extraction #
    # ---------- #

    # wavelength Step

    # ------------- #
    #               #
    # APERTURE      #
    #               #
    # ------------- #
    if args.aperture is not None and len(args.aperture) >0:
        if args.std:
            raise NotImplementedError("--std option cannot be used (yet) with --aperture. Use --auto")
        
        for target in args.aperture.split(","):
            filecubes = io.get_night_files(date, "cube.*", target=target.replace(".fits","")) if not args.inputcube else [target]
            print("cube file from which the spectra will be extracted [aperture]: "+ ", ".join(filecubes))
            
            # - loop over the file cube
            for filecube in filecubes:
                # ----------------- #
                #  Cube to Fit?     #
                # ----------------- #
                print("Automatic extraction of target %s, file: %s"%(target, filecube))
                cube_ = get_sedmcube(filecube)
                [xcentroid, ycentroid], centroids_err, position_type = \
                    astrometry.position_source(cube_, centroid=args.centroid,
                                    centroiderr=args.centroiderr,
                                    maxpos=args.maxpos)
                if args.display:
                    iplot = cube_.show(interactive=True, launch=False)
                    iplot.axim.scatter( xcentroid, ycentroid, **astrometry.MARKER_PROP[position_type] )
                    iplot.launch(vmin=args.vmin, vmax=args.vmax, notebook=False)
                    [aper_xcentroid, aper_ycentroid], [radius,
                                                       bkgd_radius] = build_aperture_param(
                        iplot._stored_picked_poly, args.ap_bkgdscale)
                else:
                    aper_xcentroid = xcentroid
                    aper_ycentroid = ycentroid
                    radius = args.buffer
                    bkgd_radius = radius * args.ap_bkgdscale

                # Actual aperture extraction
                print("INFO: Aperture extraction ongoing...")

                position_type = "aperture"
                cube_.load_adr()
                spec = cube_.get_aperture_spec(aper_xcentroid, aper_ycentroid, radius,
                                                   bkgd_annulus=[1, bkgd_radius/radius],
                                                    adr=cube_.adr)
                # --------------
                # header info passed 
                # --------------
                spec._side_properties["filename"] = cube_.filename
                for k, v in cube_.header.items():
                    if k not in spec.header:
                        spec.header.set(k, v)

                spec.header.set('POSOK', True, "Is the Target centroid inside the MLA?")
                
                spec.header.set('PYSEDMV', pysedm.__version__, "Version of pysedm used")
                spec.header.set('PYSEDMPI', "M. Rigault and D. Neill", "authors of the pysedm pipeline")
                spec.header.set('EXTRACT', "manual" if args.display else "auto", "Was the Extraction manual or automatic")
                
                spec.header.set('XPOS', xcentroid, "x centroid position at reference wavelength (in spaxels)")
                spec.header.set('YPOS', ycentroid, "y centroid position at reference wavelength (in spaxels)")
                spec.header.set('LBDAPOS', -99, "reference wavelength for the centroids (in angstrom) | not defined in apeture")
                spec.header.set('SRCPOS', position_type, "How was the centroid selected ?")

                spec.header.set("QUALITY", sedm.asses_quality(spec), "spectrum extraction quality flag [3,4 means band ; 0=default] ")
                spec.header.set("REDUCER", args.reducer, "Name of the pysedm pipeline reducer [default: auto]") ######### NOT AUTO
                
                # Aperture shape
                #fwhm_arcsec = psffit.slices[2]["slpsf"].model.fwhm * IFU_SCALE_UNIT * 2
                spec.header.set('APRAD', radius, "spaxel radius of extraction aperture")
                spec.header.set('PSFFWHM', -99, "Not defined in Aperture Mode")
                # fwhm & A/B ratio
                spec.header.set('PSFAB', -99, "A/B ratio of the PSF | Not defined in Aperture mode")

                spec.header.set('CRPIX1', 1, "")    # correct CRPIX1 from e3d
                
                spec_raw = spec.copy()
                # --------------
                # Flux Calibation
                # --------------
                # APERTURE EXTRACTION
                spec, flux_calibrated = flux_calibrate(spec, fluxcalfile=args.fluxcalsource, nofluxcal=args.nofluxcal)
                # --------------
                # Recording
                # --------------
                add_tag = "_%s"%args.tag if args.tag is not None and args.tag not in ["None", ""] else ""
                add_info_spec = "_notfluxcal" if not flux_calibrated else ""
                    
                spec_info = add_info_spec
                io._saveout_forcepsf_(filecube, cube_, cuberes=None, cubemodel=None,
                                          mode="aperture"+add_tag,spec_info=spec_info,
                                          fluxcal=flux_calibrated,
                                          cubefitted=None, spec=spec)
                # --------------
                # Recording
                # --------------
                if not args.nofig:                    
                    # SHOWING THE IFU
                    import matplotlib.pyplot as mpl
                    from matplotlib import patches
                    fig = mpl.figure(figsize=[3.5,3.5])
                    ax = fig.add_axes([0.15,0.15,0.75,0.75])
                    _ = cube_._display_im_(ax, vmax=args.vmax, vmin=args.vmin, lbdalim=[6000,9000])
                    ax.scatter(aper_xcentroid, aper_ycentroid, **astrometry.MARKER_PROP[position_type])

                    aper_circle = patches.Circle([aper_xcentroid, aper_ycentroid],
                                                     radius=radius, fc="None", ec="k",lw=1)
                    aper_back   = patches.Wedge([aper_xcentroid, aper_ycentroid], radius, 0, 360, width=radius-bkgd_radius,
                                                facecolor="0.6", edgecolor="k",
                                                linewidth=1,fill=True, alpha=0.3)
                    ax.add_patch(aper_back)
                    ax.add_patch(aper_circle)

                    ax.set_xticks(np.arange(-20,20, 5))
                    ax.set_yticks(np.arange(-20,20, 5))
                    ax.grid(color='0.6', linestyle='-', linewidth=0.5, alpha=0.5)
                    
                    ax.figure.savefig(spec.filename.replace("spec","ifu_spaxels_source").replace(".fits",".pdf"))
                    ax.figure.savefig(spec.filename.replace("spec","ifu_spaxels_source").replace(".fits",".png"), dpi=150)
                    
            # ----------------------- #                    
            # End Aperture Extraction #
            # ----------------------- #
            
    # ------------- #
    #               #
    # AUTO: PSF     #
    #               #
    # ------------- #    
    if args.auto is not None and len(args.auto) >0:
        final_slice_width = int(args.lstep)
        # - Step 1 parameters
        lbdaranges, bins = np.asarray(args.autorange.split(","), dtype="float"), int(args.autobins + 1)
        lbda_step1 = script.lbda_and_bin_to_lbda_step1(lbdaranges, bins)
        
        for target in args.auto.split(","):
            filecubes = io.get_night_files(date, "cube.*", target=target.replace(".fits","")) if not args.inputcube else [target]
            print("cube file from which the spectra will be extracted [auto]: "+ ", ".join(filecubes))
            
            # - loop over the file cube
            for filecube in filecubes:
                # ----------------- #
                #  Cube to Fit?     #
                # ----------------- #
                print("Automatic extraction of target %s, file: %s"%(target, filecube))
                cube_ = get_sedmcube(filecube)
                try:
                    [xcentroid, ycentroid], centroids_err, position_type = \
                        astrometry.position_source(cube_,
                                                   centroid=args.centroid,
                                                   centroiderr=args.centroiderr,
                                                   maxpos=args.maxpos)
                except IOError:
                    print("Astrometry file not found, using maxpos")
                    [xcentroid, ycentroid], centroids_err, position_type = \
                        astrometry.position_source(cube_,
                                                   centroid=args.centroid,
                                                   centroiderr=args.centroiderr,
                                                   maxpos=True)
                if args.display:
                    iplot = cube_.show(interactive=True, launch=False)
                    iplot.axim.scatter( xcentroid, ycentroid, **astrometry.MARKER_PROP[position_type] )
                    iplot.launch(vmin=args.vmin, vmax=args.vmax, notebook=False)
                    cube = cube_.get_partial_cube( iplot.get_selected_idx(), np.arange( len(cube_.lbda)) )
                    args.buffer = SEDM_MLA_RADIUS
                    if iplot.picked_position is not None:
                        print("You picked the position : ", iplot.picked_position )
                        print(" updating the centroid accordingly ")
                        xcentroid, ycentroid = iplot.picked_position
                        centroids_err = [2., 2.]
                        position_type = "manual"
                else:
                    cube = cube_

                    
                # Centroid ?
                print("INFO: PSF centroid (%s)**"%position_type)
                print("centroid: %.1f %.1f"%(xcentroid, ycentroid)+ " error: %.1f %.1f"%(centroids_err[0], centroids_err[1]))
                if not geometry.Point(0,0).buffer(sedm.SEDM_MLA_RADIUS).contains(geometry.Point(xcentroid, ycentroid)):
                    print("WARNING: centroid outside the MLA field of view. NaN spectra returned. * Quality 3. *")
                    spec  = get_spectrum( sedm.SEDM_LBDA, np.ones(len(sedm.SEDM_LBDA))*np.NaN, header=cube.header )
                    # Other
                    cubemodel, psfmodel, bkgdmodel, psffit, slpsf = None, None, None, None, None
                    cube_to_fit = None
                    # Header info
                    POSOK = False
                    lbdaref     = "nan"
                    fwhm_arcsec = "nan"
                    psf_ab     = "nan"
                    psf_pa      = "nan"
                    psf_airmass = "nan"
                    psf_chi2    = "nan"                    
                else:
                    # Aperture area ?
                    point_polygon = geometry.Point(xcentroid, ycentroid).buffer( float(args.buffer) )
                    
                    # => Cube to fit
                    cube_to_fit = cube.get_partial_cube( cube.get_spaxels_within_polygon(point_polygon),
                                                      np.arange(len(cube.lbda)))
                    # --------------
                    # Fitting
                    # --------------
                    print("INFO: Starting MetaSlice fit")
                    spec, cubemodel, psfmodel, bkgdmodel, psffit, slpsf  = \
                      script.extract_star(cube_to_fit, lbda_step1,
                                          centroids=[xcentroid, ycentroid], centroids_err=centroids_err,
                                          spaxel_unit = IFU_SCALE_UNIT,
                                          final_slice_width = final_slice_width,
                                          psfmodel=args.psfmodel, normalized=args.normed)
                    # Hack to be removed:
                    #print("INFO: Temporary variance hacking to be removed ")
                    spec._properties['rawvariance'] = np.ones(len(spec.lbda)) * np.nanmedian( spec.variance[spec.variance>0] )

                    # Divide out exposure time
                    expt = spec.header.get("EXPTIME", 1.0)
                    print("Dividing counts by %s seconds" % expt)
                    spec.scale_by(expt)
                    spec.header.set("CALSCL", True, "Exposure time divided out")

                    if final_slice_width != 1:
                        spec = spec.reshape(cube.lbda)
                    # For header:
                    POSOK       = True
                    lbdaref     = psffit.adrfitter.model.lbdaref
                    fwhm_arcsec = psffit.slices[2]["slpsf"].model.fwhm * IFU_SCALE_UNIT * 2
                    psf_ab     = psffit.slices[2]["slpsf"].fitvalues['ab']
                    psf_pa      = psffit.adrfitter.fitvalues["parangle"]
                    psf_airmass = psffit.adrfitter.fitvalues["airmass"]
                    psf_chi2    = psffit.adrfitter.fitvalues["chi2"]/psffit.adrfitter.dof
                    
                # --------------
                # header info passed
                # --------------
                spec._side_properties["filename"] = cube_.filename
                for k,v in cube.header.items():
                    if k not in spec.header:
                        spec.header.set(k,v)
                # Additional information
                # centroid
                spec.header.set('POSOK', POSOK, "Is the Target centroid inside the MLA?")
                spec.header.set('PYSEDMV', pysedm.__version__, "Version of pysedm used")
                spec.header.set('PYSEDMPI', "M. Rigault and D. Neill", "authors of the pysedm pipeline")
                spec.header.set('PSFV', psfcube.__version__, "Version of psfcube used")
                spec.header.set('PSFPI', "M. Rigault", "authors of the psfcube")
                spec.header.set('PSFMODEL', args.psfmodel, "PSF model used in psfcube")
                spec.header.set('EXTRACT', "manual" if args.display else "auto", "Was the Extraction manual or automatic")
                
                spec.header.set('XPOS', xcentroid, "x centroid position at reference wavelength (in spaxels)")
                spec.header.set('YPOS', ycentroid, "y centroid position at reference wavelength (in spaxels)")
                spec.header.set('LBDAPOS',lbdaref , "reference wavelength for the centroids (in angstrom)")
                spec.header.set('SRCPOS', position_type, "How was the centroid selected ?")
                
                # PSF shape
                
                spec.header.set('PSFFWHM', fwhm_arcsec, "twice the radius needed to reach half of the pick brightness [in arcsec]")
                
                # fwhm & A/B ratio
                spec.header.set('PSFAB', psf_ab , "A/B ratio of the PSF")
                
                # ADR
                spec.header.set('PSFADRPA', psf_pa, "Fitted ADR paralactic angle")
                spec.header.set('PSFADRZ', psf_airmass, "Fitted ADR airmass")
                try:
                    spec.header.set('PSFADRC2',psf_chi2, "ADR chi2/dof")
                except:
                    spec.header.set('PSFADRC2', "nan", "ADR chi2/dof")

                spec.header.set("QUALITY", sedm.asses_quality(spec), "spectrum extraction quality flag [3,4 means band ; 0=default] ")
                spec.header.set("REDUCER", args.reducer, "Name of the pysedm pipeline reducer [default: auto]") ######### NOT AUTO

                spec.header.set('CRPIX1', 1, "")    # correct CRPIX1 from e3d
                
                # Basic quality check ?

                spec_raw = spec.copy()
                # --------------
                # Flux Calibation
                # --------------
                # PSF EXTRACTION
                spec, flux_calibrated  = flux_calibrate(spec, fluxcalfile=args.fluxcalsource, nofluxcal=args.nofluxcal)
                # -------------
                # Cut Edges of fluxcalibrated
                # -------------
                if flux_calibrated and not args.notroncation:
                    from pysedm.sedm import LBDA_PIXEL_CUT
                    spec = get_spectrum(spec.lbda[LBDA_PIXEL_CUT:-LBDA_PIXEL_CUT], spec.data[LBDA_PIXEL_CUT:-LBDA_PIXEL_CUT],
                                        variance=spec.variance[LBDA_PIXEL_CUT:-LBDA_PIXEL_CUT] if spec.has_variance() else None,
                                        header=spec.header, logwave=spec.spec_prop["logwave"])
                    spec._side_properties["filename"] = spec_raw.filename
                    spec.header.set("EDGECUT", True, "have some edge pixels been removed during flux cal")
                    spec.header.set("EDGECUTL", LBDA_PIXEL_CUT, "number of edge pixels removed during flux cal")
                else:
                    spec.header.set("EDGECUT", False, "bluer and redder pixel of the spectra have been removed during flux calibration")
                    spec.header.set("EDGECUTL", None, "number of bluer and redder pixels of the spectra have been removed during flux calibration")
                                        
                # --------------
                # Recording
                # --------------
                add_tag = "_%s"%args.tag if args.tag is not None and args.tag not in ["None", ""] else ""
                add_info_spec = "_notfluxcal" if not flux_calibrated else ""
                spec_info = "_lstep%s"%final_slice_width + add_info_spec
                # MAIN IO
                io._saveout_forcepsf_(filecube, cube, cuberes=None, cubemodel=cubemodel,
                                      mode="auto"+add_tag,spec_info=spec_info, fluxcal=flux_calibrated,
                                      cubefitted=cube_to_fit, spec=spec)
                # Figure
                if not args.nofig:
                    import matplotlib.pyplot as mpl
                    # Pure spaxel
                    fig = mpl.figure(figsize=[3.5,3.5])
                    ax = fig.add_axes([0.15,0.15,0.75,0.75])
                    _ = cube_._display_im_(ax, vmax=args.vmax, vmin=args.vmin, lbdalim=[6000,9000])
                    if POSOK:
                        x,y = np.asarray(cube_to_fit.index_to_xy(cube_to_fit.indexes)).T
                        ax.plot(x, y, marker=".", ls="None", ms=1, color="k")
                        ax.scatter(xcentroid, ycentroid, **astrometry.MARKER_PROP[position_type])
                    else:
                        ax.text(0.5,0.95, "Target outside the MLA \n [%.1f, %.1f] (in spaxels)"%(xcentroid, ycentroid),
                                    fontsize="large", color="k",backgroundcolor=mpl.cm.binary(0.1,0.4),
                                    transform=ax.transAxes, va="top", ha="center")
                        
                    ax.set_xticks(np.arange(-20,20, 5))
                    ax.set_yticks(np.arange(-20,20, 5))

                    ax.grid(color='0.6', linestyle='-', linewidth=0.5, alpha=0.5)
                    
                    ax.figure.savefig(spec.filename.replace("spec","ifu_spaxels_source").replace(".fits",".pdf"))
                    ax.figure.savefig(spec.filename.replace("spec","ifu_spaxels_source").replace(".fits",".png"), dpi=150)
                    
                    if psffit is not None:
                        x,y = np.asarray(cube_to_fit.index_to_xy(cube_to_fit.indexes)).T
                        # SHOWING ADR
                        psffit.show_adr(savefile=spec.filename.replace("spec","adr_fit").replace(".fits",".pdf") )
                        psffit.show_adr(savefile=spec.filename.replace("spec","adr_fit").replace(".fits",".png") )
                        # SHOWING PSFPROFILE (metaslice)
                        psffit.slices[2]["slpsf"].show(savefile=spec.filename.replace("spec","psfprofile").replace(".fits",".pdf"))
                        psffit.slices[2]["slpsf"].show(savefile=spec.filename.replace("spec","psfprofile").replace(".fits",".png"))
                        
                        # SHOWING SPAXEL (IFU)
                        
                        cube_.show(show=False)
                        ax = mpl.gca()
                        ax.plot(x,y, marker=".", ls="None", ms=1, color="k")
                        ax.scatter(xcentroid, ycentroid, **astrometry.MARKER_PROP[position_type])
                        ax.set_xticks(np.arange(-20,20, 5))
                        ax.set_yticks(np.arange(-20,20, 5))
                        ax.grid(color='0.6', linestyle='-', linewidth=0.5, alpha=0.5)
                    
                        ax.figure.savefig(spec.filename.replace("spec","spaxels_source").replace(".fits",".pdf"))
                    
                    
                    # Special Standard
                    if cube.header['IMGTYPE'].lower() in ['standard'] and flux_calibrated:
                        from pysedm.fluxcalibration import show_fluxcalibrated_standard
                        show_fluxcalibrated_standard(spec, savefile=spec.filename.replace("spec","calibcheck_spec").replace(".fits",".pdf"))
                        show_fluxcalibrated_standard(spec, savefile=spec.filename.replace("spec","calibcheck_spec").replace(".fits",".png"))
                        
                # -----------------
                #  Is that a STD  ?
                # -----------------
                if args.std and cube.header['IMGTYPE'].lower() in ['standard'] and 'AIRMASS' in cube.header:
                    if spec_raw.header['QUALITY'] == 0:
                        # Based on the flux non calibrated spectsra
                        spec_raw.header['OBJECT'] = cube.header['OBJECT']
                        for k,v in cube.header.items():
                            if k not in spec_raw.header:
                                spec_raw.header.set(k, v)

                        try:
                            speccal, fl = fluxcalibration.get_fluxcalibrator(
                                spec_raw, fullout=True)
                            speccal.header["SOURCE"] = (
                                spec.filename.split("/")[-1],
                                "This object has been derived from this file")
                            speccal.header["PYSEDMT"] = (
                                "Flux Calibration Spectrum",
                                "Object to use to flux calibrate")
                            filename_inv = spec.filename.replace(
                                io.PROD_SPECROOT,
                                io.PROD_SENSITIVITYROOT).replace("notfluxcal",
                                                                 "")
                            speccal._side_properties['filename'] = filename_inv
                            speccal.writeto(filename_inv)
                            if not args.nofig:
                                fl.show(
                                    savefile=speccal.filename.replace(".fits",
                                                                      ".pdf"),
                                    show=False)
                                fl.show(
                                    savefile=speccal.filename.replace(".fits",
                                                                      ".png"),
                                    show=False)
                        except OSError:
                            print("WARNING: no reference spectrum for target, "
                                  "skipping flux calibration")



                    else:
                        print("WARNING: Standard spectrum of low quality, "
                              "skipping fluxcal generation")
                # - for the record
                extracted_objects.append(spec)
                
            # -------------------------- #                    
            # End Auto (PSF) Extraction  #
            # -------------------------- #



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import warnings
import numpy         as np
import matplotlib.pyplot as mpl
from .tools import make_method
from pyifu.mplinteractive import InteractiveCube
LAMPS = np.asarray(["Cd","Hg","Xe"])

try:
    CMAP = mpl.cm.viridis
except:
    warnings.warn("You should update your matplotlib. viridis does not exist...")
    CMAP = mpl.cm.PuBu



__all__ = ["get_lamp_color", "figout"]
    
def get_lamp_color(lamp, alpha):
    """ """
    i = float(np.argwhere(LAMPS == lamp)[0][0])/len(LAMPS)
    return CMAP(i, alpha)


@make_method(mpl.Figure)
def figout(fig,savefile=None,show=True,add_thumbnails=False,
           dpi=200):
    """This methods parse the show/savefile to know if the figure
    shall the shown or saved."""
    
    if savefile in ["dont_show","_dont_show_","_do_not_show_"]:
        show = False
        savefile = None

    if savefile is not None:
        if not savefile.endswith(".pdf"):
            extention = ".png" if not savefile.endswith(".png") else ""
            fig.savefig(savefile+extention,dpi=dpi)
            
        if not savefile.endswith(".png"):
            extention = ".pdf" if not savefile.endswith(".pdf") else ""
            fig.savefig(savefile+extention)
            
        if add_thumbnails:
            fig.savefig(savefile+"_thumb"+'.png',dpi=dpi/10.)
            
    elif show:
        fig.canvas.draw()
        fig.show()

@make_method(mpl.Axes)        
def set_axes_edgecolor(ax, color,  ticks=True, labels=False):
    """ """
    import matplotlib
    prop = {}
    if ticks:
        prop["color"] = color
        prop["which"] = "both"
    if labels:
        prop["labelcolor"] = color
    ax.tick_params(**prop)
    for child in ax.get_children():
        if isinstance(child, matplotlib.spines.Spine):
            child.set_color(color)

###############################
#                             #
# SEDMachine Plotting special #
#                             #
###############################
class InteractiveCubeandCCD( InteractiveCube ):
    PROPERTIES = ["ccd", "axccd"]

    def launch(self, *args, **kwargs):
        """ """
        
        self._ccdshow = self.ccdimage.show(ax = self.axccd, aspect="auto", show=False)["imshow"]
        super(InteractiveCubeandCCD, self).launch(*args, **kwargs)

    def set_to_origin(self):
        """ Defines the default parameters """
        super(InteractiveCubeandCCD, self).set_to_origin()
        self._trace_patches = []

    def reset(self):
        """ Set back everything to origin """
        super(InteractiveCubeandCCD, self).reset()
        self.clean_axccd(update_limits=True)
        
    # =========== #
    #  SETTER     #
    # =========== #
    def set_ccd(self, ccd):
        """ """
        self._properties["ccd"] = ccd
        
    def set_figure(self, fig=None, **kwargs):
        """ attach a figure to this method. """
        if fig is None:
            figprop = kwargs_update(dict(figsize=[10,7]), **kwargs)
            self._properties["figure"] = mpl.figure(**figprop)
        elif matplotlib.figure.Figure not in fig.__class__.__mro__:
            raise TypeError("The given fig must be a matplotlib.figure.Figure (or child of)")
        else:
            self._properties["figure"] = mpl.figure(**figprop)

    def set_axes(self, axes=None, **kwargs):
        """ """
        if axes is None and not self.has_figure():
            raise AttributeError("If no axes given, please first set the figure.")
        
        if axes is None:
            figsizes = self.fig.get_size_inches()
            axspec = self.fig.add_axes([0.10,0.6,0.5,0.35])
            axim   = self.fig.add_axes([0.65,0.6,0.75*(figsizes[1]*0.5)/float(figsizes[0]),0.35])

            axccd  = self.fig.add_axes([0.10,0.10,axim.get_position().xmax- 0.1, 0.4])
            
            axspec.set_xlabel(r"Wavelength", fontsize="large")
            axspec.set_ylabel(r"Flux", fontsize="large")
            self.set_axes([axspec,axim,axccd])
            
        elif len(axes) != 3:
            raise TypeError("you must provide 2 axes [axspec and axim] and both have to be matplotlib.axes(._axes).Axes (or child of)")
        else:
            # - actual setting
            self._properties["axspec"], self._properties["axim"], self._properties["axccd"] = axes
            if not self.has_figure():
                self.set_figure(self.axspec.figure)

    # ------------- #
    #  Show Things  #
    # ------------- #
    def show_picked_traces(self):
        """ """
        if not self._hold:
            self.clean_axccd()
            
        self._trace_patches = self.ccdimage.display_traces(self.axccd, self.get_selected_idx(),
                                                            facecolors="None", edgecolors=self._active_color)

    def clean_axccd(self, update_limits=False):
        """ """
        self.axccd.patches = []
        if update_limits:
            self.axccd.set_xlim(0, self.ccdimage.shape[0])
            self.axccd.set_ylim(0, self.ccdimage.shape[1])
        
    # ================= #
    # Change For CCD    #
    # ================= #
    def update_figure_fromaxim(self):
        """ What would happen once the spaxels are picked. """
        if len(self.selected_spaxels)>0:
            self.show_picked_spaxels()
            self.show_picked_spectrum()
            self.show_picked_traces()
            self.fig.canvas.draw()

    # ================= #
    # Properties        #
    # ================= #
    @property
    def axccd(self):
        """ """
        return self._properties["axccd"]
    
    @property
    def ccdimage(self):
        """ """
        return self._properties["ccd"]


